import itertools
import unittest
from copy import copy

import numpy as np
import pytest

from coremltools._deps import _HAS_KERAS2_TF, _HAS_KERAS_TF
from coremltools.models.utils import _macos_version, _is_macos

np.random.seed(1377)

if _HAS_KERAS2_TF or _HAS_KERAS_TF:
    import keras
    from keras.models import Sequential
    from keras.layers import LSTM, GRU, SimpleRNN, RepeatVector
    from keras.layers.wrappers import Bidirectional
    import keras.backend as K
    from coremltools.converters import keras as keras_converter

"""
=============================
Utility Functions
=============================
"""


def get_recurrent_activation_name_from_keras(activation):
    if activation == keras.activations.sigmoid:
        activation_str = "SIGMOID"
    elif activation == keras.activations.hard_sigmoid:
        activation_str = "SIGMOID_HARD"
    elif activation == keras.activations.tanh:
        activation_str = "TANH"
    elif activation == keras.activations.relu:
        activation_str = "RELU"
    elif activation == keras.activations.linear:
        activation_str = "LINEAR"
    else:
        raise NotImplementedError(
            "activation %s not supported for Recurrent layer." % activation
        )
    return activation_str


def linear(x, alpha=1, beta=0):
    return alpha * x + beta


def relu(x):
    return np.maximum(0, x)


def sigmoid(x):
    return 1.0 / (1 + np.exp(-x))


def hard_sigmoid(x, alpha=0.2, beta=0.5):
    return np.minimum(np.maximum(alpha * x + beta, 0), 1)


def tanh(x):
    return np.tanh(x)


def apply_act(x, option):
    if option == "TANH":
        return tanh(x)
    elif option == "RELU":
        return relu(x)
    elif option == "SIGMOID":
        return sigmoid(x)
    elif option == "SIGMOID_HARD":
        return hard_sigmoid(x)
    elif option == "LINEAR":
        return linear(x)


def clip(x, threshold=50.0):
    return np.maximum(np.minimum(x, threshold), -threshold)


def valid_params(params):
    """Checks if this combination of parameters is allowed by Keras"""
    return not (params["input_dims"][1] == 1 and params["unroll"])


def _compute_SNR(x, y):
    x = x.flatten()
    y = y.flatten()
    noise = x - y
    noise_var = np.sum(noise ** 2) / len(noise) + 1e-7
    signal_energy = np.sum(y ** 2) / len(y)
    signal_energy2 = np.sum(x ** 2) / len(x)
    if signal_energy < 1e-5 and signal_energy2 < 1e-5:
        return 50, 50, 0
    max_signal_energy = np.amax(y ** 2)
    SNR = 10 * np.log10(signal_energy / noise_var)
    PSNR = 10 * np.log10(max_signal_energy / noise_var)
    return SNR, PSNR, signal_energy


"""
=============================
Numpy implementations
=============================
"""


def get_numpy_prediction_gru(model, X):
    X = X[0, :, :]
    seq_len, input_size = X.shape
    keras_layer = model.layers[0]
    return_seq = keras_layer.return_sequences
    if keras_layer.go_backwards:
        X = X[::-1, :]

    if _HAS_KERAS2_TF:
        hidden_size = keras_layer.units

        keras_W_h = keras_layer.get_weights()[1].T
        R_z = keras_W_h[0 * hidden_size :][:hidden_size]
        R_r = keras_W_h[1 * hidden_size :][:hidden_size]
        R_o = keras_W_h[2 * hidden_size :][:hidden_size]

        keras_W_x = keras_layer.get_weights()[0].T
        W_z = keras_W_x[0 * hidden_size :][:hidden_size]
        W_r = keras_W_x[1 * hidden_size :][:hidden_size]
        W_o = keras_W_x[2 * hidden_size :][:hidden_size]

        keras_b = keras_layer.get_weights()[2]
        b_z = keras_b[0 * hidden_size :][:hidden_size]
        b_r = keras_b[1 * hidden_size :][:hidden_size]
        b_o = keras_b[2 * hidden_size :][:hidden_size]

        inner_activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.recurrent_activation
        )
        activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.activation
        )

    else:
        hidden_size = keras_layer.output_dim

        W_z = keras_layer.get_weights()[0].T
        W_r = keras_layer.get_weights()[3].T
        W_o = keras_layer.get_weights()[6].T

        R_z = keras_layer.get_weights()[1].T
        R_r = keras_layer.get_weights()[4].T
        R_o = keras_layer.get_weights()[7].T

        b_z = keras_layer.get_weights()[2]
        b_r = keras_layer.get_weights()[5]
        b_o = keras_layer.get_weights()[8]

        inner_activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.inner_activation
        )
        activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.activation
        )

    h = np.zeros((hidden_size))
    c = np.zeros((hidden_size))
    np_out = np.zeros((seq_len, hidden_size))
    for k in range(seq_len):
        x = X[k, :]
        z = apply_act(clip(np.dot(W_z, x) + np.dot(R_z, h) + b_z), inner_activation_str)
        r = apply_act(clip(np.dot(W_r, x) + np.dot(R_r, h) + b_r), inner_activation_str)
        c = clip(h * r)
        o = apply_act(clip(np.dot(W_o, x) + np.dot(R_o, c) + b_o), activation_str)
        h = (1 - z) * o + z * h
        np_out[k, :] = h

    if return_seq:
        np_out_final = np_out
    else:
        np_out_final = np_out[-1, :]
    return np_out_final


def get_numpy_prediction_unilstm(model, X):
    X = X[0, :, :]
    seq_len, input_size = X.shape
    keras_layer = model.layers[0]
    return_seq = keras_layer.return_sequences
    if keras_layer.go_backwards:
        X = X[::-1, :]

    if _HAS_KERAS2_TF:
        hidden_size = keras_layer.units

        keras_W_h = keras_layer.get_weights()[1].T
        R_i = keras_W_h[0 * hidden_size :][:hidden_size]
        R_f = keras_W_h[1 * hidden_size :][:hidden_size]
        R_o = keras_W_h[3 * hidden_size :][:hidden_size]
        R_g = keras_W_h[2 * hidden_size :][:hidden_size]

        keras_W_x = keras_layer.get_weights()[0].T
        W_i = keras_W_x[0 * hidden_size :][:hidden_size]
        W_f = keras_W_x[1 * hidden_size :][:hidden_size]
        W_o = keras_W_x[3 * hidden_size :][:hidden_size]
        W_g = keras_W_x[2 * hidden_size :][:hidden_size]

        keras_b = keras_layer.get_weights()[2]
        b_i = keras_b[0 * hidden_size :][:hidden_size]
        b_f = keras_b[1 * hidden_size :][:hidden_size]
        b_o = keras_b[3 * hidden_size :][:hidden_size]
        b_g = keras_b[2 * hidden_size :][:hidden_size]

        inner_activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.recurrent_activation
        )
        activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.activation
        )

    else:
        hidden_size = keras_layer.output_dim

        R_i = keras_layer.get_weights()[1].T
        R_f = keras_layer.get_weights()[7].T
        R_o = keras_layer.get_weights()[10].T
        R_g = keras_layer.get_weights()[4].T

        W_i = keras_layer.get_weights()[0].T
        W_f = keras_layer.get_weights()[6].T
        W_o = keras_layer.get_weights()[9].T
        W_g = keras_layer.get_weights()[3].T

        b_i = keras_layer.get_weights()[2]
        b_f = keras_layer.get_weights()[8]
        b_o = keras_layer.get_weights()[11]
        b_g = keras_layer.get_weights()[5]

        inner_activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.inner_activation
        )
        activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.activation
        )

    h = np.zeros((hidden_size))
    c = np.zeros((hidden_size))
    np_out = np.zeros((seq_len, hidden_size))
    for k in range(seq_len):
        x = X[k, :]
        i = apply_act(clip(np.dot(W_i, x) + np.dot(R_i, h) + b_i), inner_activation_str)
        f = apply_act(clip(np.dot(W_f, x) + np.dot(R_f, h) + b_f), inner_activation_str)
        g = apply_act(clip(np.dot(W_g, x) + np.dot(R_g, h) + b_g), activation_str)
        c = c * f + i * g
        c = clip(c, 50000.0)
        o = apply_act(clip(np.dot(W_o, x) + np.dot(R_o, h) + b_o), inner_activation_str)
        h = o * apply_act(c, activation_str)
        np_out[k, :] = h
    if return_seq:
        np_out_final = np_out
    else:
        np_out_final = np_out[-1, :]
    return np_out_final


def get_numpy_prediction_bilstm_batched(model, X):
    batch, _, _ = X.shape
    out = []
    for i in range(batch):
        out.append(
            get_numpy_prediction_bilstm(model, np.expand_dims(X[i, :, :], axis=0))
        )
    return np.stack(out, axis=0)


def get_numpy_prediction_bilstm(model, X):
    X = X[0, :, :]
    seq_len, input_size = X.shape
    keras_layer = model.layers[0]
    return_seq = keras_layer.return_sequences

    if _HAS_KERAS2_TF:
        hidden_size = keras_layer.forward_layer.units

        keras_W_h = keras_layer.forward_layer.get_weights()[1].T
        R_i = keras_W_h[0 * hidden_size :][:hidden_size]
        R_f = keras_W_h[1 * hidden_size :][:hidden_size]
        R_o = keras_W_h[3 * hidden_size :][:hidden_size]
        R_g = keras_W_h[2 * hidden_size :][:hidden_size]

        keras_W_x = keras_layer.forward_layer.get_weights()[0].T
        W_i = keras_W_x[0 * hidden_size :][:hidden_size]
        W_f = keras_W_x[1 * hidden_size :][:hidden_size]
        W_o = keras_W_x[3 * hidden_size :][:hidden_size]
        W_g = keras_W_x[2 * hidden_size :][:hidden_size]

        keras_b = keras_layer.forward_layer.get_weights()[2]
        b_i = keras_b[0 * hidden_size :][:hidden_size]
        b_f = keras_b[1 * hidden_size :][:hidden_size]
        b_o = keras_b[3 * hidden_size :][:hidden_size]
        b_g = keras_b[2 * hidden_size :][:hidden_size]

        keras_W_h = keras_layer.backward_layer.get_weights()[1].T
        R_i_back = keras_W_h[0 * hidden_size :][:hidden_size]
        R_f_back = keras_W_h[1 * hidden_size :][:hidden_size]
        R_o_back = keras_W_h[3 * hidden_size :][:hidden_size]
        R_g_back = keras_W_h[2 * hidden_size :][:hidden_size]

        keras_W_x = keras_layer.backward_layer.get_weights()[0].T
        W_i_back = keras_W_x[0 * hidden_size :][:hidden_size]
        W_f_back = keras_W_x[1 * hidden_size :][:hidden_size]
        W_o_back = keras_W_x[3 * hidden_size :][:hidden_size]
        W_g_back = keras_W_x[2 * hidden_size :][:hidden_size]

        keras_b = keras_layer.backward_layer.get_weights()[2]
        b_i_back = keras_b[0 * hidden_size :][:hidden_size]
        b_f_back = keras_b[1 * hidden_size :][:hidden_size]
        b_o_back = keras_b[3 * hidden_size :][:hidden_size]
        b_g_back = keras_b[2 * hidden_size :][:hidden_size]

        inner_activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.forward_layer.recurrent_activation
        )
        activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.forward_layer.activation
        )

    else:
        hidden_size = keras_layer.forward_layer.output_dim

        R_i = keras_layer.get_weights()[1].T
        R_f = keras_layer.get_weights()[7].T
        R_o = keras_layer.get_weights()[10].T
        R_g = keras_layer.get_weights()[4].T

        W_i = keras_layer.get_weights()[0].T
        W_f = keras_layer.get_weights()[6].T
        W_o = keras_layer.get_weights()[9].T
        W_g = keras_layer.get_weights()[3].T

        b_i = keras_layer.get_weights()[2]
        b_f = keras_layer.get_weights()[8]
        b_o = keras_layer.get_weights()[11]
        b_g = keras_layer.get_weights()[5]

        R_i_back = keras_layer.backward_layer.get_weights()[1].T
        R_f_back = keras_layer.backward_layer.get_weights()[7].T
        R_o_back = keras_layer.backward_layer.get_weights()[10].T
        R_g_back = keras_layer.backward_layer.get_weights()[4].T

        W_i_back = keras_layer.backward_layer.get_weights()[0].T
        W_f_back = keras_layer.backward_layer.get_weights()[6].T
        W_o_back = keras_layer.backward_layer.get_weights()[9].T
        W_g_back = keras_layer.backward_layer.get_weights()[3].T

        b_i_back = keras_layer.backward_layer.get_weights()[2]
        b_f_back = keras_layer.backward_layer.get_weights()[8]
        b_o_back = keras_layer.backward_layer.get_weights()[11]
        b_g_back = keras_layer.backward_layer.get_weights()[5]

        inner_activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.forward_layer.inner_activation
        )
        activation_str = get_recurrent_activation_name_from_keras(
            keras_layer.forward_layer.activation
        )

    h = np.zeros((hidden_size))
    c = np.zeros((hidden_size))
    np_out_forward = np.zeros((seq_len, hidden_size))
    for k in range(seq_len):
        x = X[k, :]
        i = apply_act(clip(np.dot(W_i, x) + np.dot(R_i, h) + b_i), inner_activation_str)
        f = apply_act(clip(np.dot(W_f, x) + np.dot(R_f, h) + b_f), inner_activation_str)
        g = apply_act(clip(np.dot(W_g, x) + np.dot(R_g, h) + b_g), activation_str)
        c = c * f + i * g
        c = clip(c, 50000.0)
        o = apply_act(clip(np.dot(W_o, x) + np.dot(R_o, h) + b_o), inner_activation_str)
        h = o * apply_act(c, activation_str)
        np_out_forward[k, :] = h

    h = np.zeros((hidden_size))
    c = np.zeros((hidden_size))
    np_out_backward = np.zeros((seq_len, hidden_size))
    for k in range(seq_len):
        x = X[seq_len - k - 1, :]
        i = apply_act(
            clip(np.dot(W_i_back, x) + np.dot(R_i_back, h) + b_i_back),
            inner_activation_str,
        )
        f = apply_act(
            clip(np.dot(W_f_back, x) + np.dot(R_f_back, h) + b_f_back),
            inner_activation_str,
        )
        g = apply_act(
            clip(np.dot(W_g_back, x) + np.dot(R_g_back, h) + b_g_back), activation_str
        )
        c = c * f + i * g
        c = clip(c, 50000.0)
        o = apply_act(
            clip(np.dot(W_o_back, x) + np.dot(R_o_back, h) + b_o_back),
            inner_activation_str,
        )
        h = o * apply_act(c, activation_str)
        np_out_backward[k, :] = h

    if return_seq:
        np_out_final = np.zeros((seq_len, 2 * hidden_size))
        for k in range(seq_len):
            np_out_final[k, :hidden_size] = np_out_forward[k, :]
            np_out_final[k, hidden_size:] = np_out_backward[seq_len - k - 1, :]
    else:
        np_out_final = np.zeros((2 * hidden_size))
        np_out_final[:hidden_size] = np_out_forward[-1, :]
        np_out_final[hidden_size:] = np_out_backward[-1, :]
    return np_out_final


"""
=============================
Nosetest Functions
=============================
"""


def get_mlkit_model_from_path(model):
    from coremltools.converters import keras as keras_converter

    model = keras_converter.convert(model, ["data"], ["output"])
    return model


def generate_input(dim0, dim1, dim2):
    input_data = np.random.rand(dim0, dim1, dim2).astype(
        "f"
    )  # astype() should be removed
    return input_data


def simple_model_eval(params, model):
    mlkitmodel = get_mlkit_model_from_path(model)
    # New test case takes in 2D input as opposed to uniform 3d input across all other tests
    if len(params[0]["input_dims"]) == 3:
        input_data = generate_input(
            params[0]["input_dims"][0],
            params[0]["input_dims"][1],
            params[0]["input_dims"][2],
        )
        keras_preds = model.predict(input_data).flatten()
    elif len(params[0]["input_dims"]) == 2:
        input_data = np.squeeze(
            np.random.rand(params[0]["input_dims"][0], params[0]["input_dims"][1])
        )
        keras_preds = model.predict(
            input_data.reshape((params[0]["input_dims"][0], params[0]["input_dims"][1]))
        ).flatten()
    if len(params[0]["input_dims"]) == 3:
        input_data = np.transpose(input_data, [1, 0, 2])
    if _is_macos() and _macos_version() >= (10, 13):
        coreml_preds = mlkitmodel.predict({"data": input_data})["output"].flatten()
        if K.tensorflow_backend._SESSION:
            import tensorflow as tf

            tf.reset_default_graph()
            K.tensorflow_backend._SESSION.close()
            K.tensorflow_backend._SESSION = None

        max_denominator = np.maximum(
            np.maximum(np.abs(coreml_preds), np.abs(keras_preds)), 1.0
        )
        relative_error = coreml_preds / max_denominator - keras_preds / max_denominator
        return relative_error, keras_preds, coreml_preds
    else:
        return [], None, None


class SimpleTestCase(unittest.TestCase):
    """
    Test Simple test cases to make sure layers work under basic params. Also, template for testing
    different failing test cases from stress tests
    """

    def _test_simple_rnn(self, keras_major_version):
        params = (
            dict(
                input_dims=[1, 2, 100],
                go_backwards=False,
                activation="tanh",
                stateful=False,
                unroll=False,
                return_sequences=True,
                output_dim=4,  # Passes for < 3
            ),
        )
        model = Sequential()
        if keras_major_version == 2:
            model.add(
                SimpleRNN(
                    units=params[0]["output_dim"],
                    input_shape=(
                        params[0]["input_dims"][1],
                        params[0]["input_dims"][2],
                    ),
                    activation=params[0]["activation"],
                    return_sequences=params[0]["return_sequences"],
                    go_backwards=params[0]["go_backwards"],
                    unroll=True,
                )
            )
        else:
            model.add(
                SimpleRNN(
                    output_dim=params[0]["output_dim"],
                    input_length=params[0]["input_dims"][1],
                    input_dim=params[0]["input_dims"][2],
                    activation=params[0]["activation"],
                    return_sequences=params[0]["return_sequences"],
                    go_backwards=params[0]["go_backwards"],
                    unroll=True,
                )
            )
        relative_error, keras_preds, coreml_preds = simple_model_eval(params, model)
        for i in range(len(relative_error)):
            self.assertLessEqual(relative_error[i], 0.01)

    def _test_simple_lstm(self, keras_major_version):
        params = (
            dict(
                input_dims=[1, 3, 5],
                go_backwards=True,
                activation="linear",
                stateful=False,
                unroll=False,
                return_sequences=False,
                output_dim=3,
                inner_activation="linear",
            ),
        )
        model = Sequential()
        if keras_major_version == 2:
            model.add(
                LSTM(
                    units=params[0]["output_dim"],
                    input_shape=(
                        params[0]["input_dims"][1],
                        params[0]["input_dims"][2],
                    ),
                    activation=params[0]["activation"],
                    return_sequences=params[0]["return_sequences"],
                    go_backwards=params[0]["go_backwards"],
                    unroll=True,
                    recurrent_activation="linear",
                )
            )
        else:
            model.add(
                LSTM(
                    output_dim=params[0]["output_dim"],
                    input_length=params[0]["input_dims"][1],
                    input_dim=params[0]["input_dims"][2],
                    activation=params[0]["activation"],
                    return_sequences=params[0]["return_sequences"],
                    go_backwards=params[0]["go_backwards"],
                    unroll=True,
                    inner_activation="linear",
                )
            )
        relative_error, keras_preds, coreml_preds = simple_model_eval(params, model)
        for i in range(len(relative_error)):
            self.assertLessEqual(relative_error[i], 0.01)

    def _test_simple_gru(self, keras_major_version):
        params = (
            dict(
                input_dims=[1, 4, 8],
                go_backwards=False,
                activation="tanh",
                stateful=False,
                unroll=False,
                return_sequences=False,
                output_dim=4,
            ),
        )
        model = Sequential()
        if keras_major_version == 2:
            model.add(
                GRU(
                    units=params[0]["output_dim"],
                    input_shape=(
                        params[0]["input_dims"][1],
                        params[0]["input_dims"][2],
                    ),
                    activation=params[0]["activation"],
                    recurrent_activation="sigmoid",
                    return_sequences=params[0]["return_sequences"],
                    go_backwards=params[0]["go_backwards"],
                    unroll=True,
                )
            )
        else:
            model.add(
                GRU(
                    output_dim=params[0]["output_dim"],
                    input_length=params[0]["input_dims"][1],
                    input_dim=params[0]["input_dims"][2],
                    activation=params[0]["activation"],
                    inner_activation="sigmoid",
                    return_sequences=params[0]["return_sequences"],
                    go_backwards=params[0]["go_backwards"],
                    unroll=True,
                )
            )
        model.set_weights([np.random.rand(*w.shape) for w in model.get_weights()])
        relative_error, keras_preds, coreml_preds = simple_model_eval(params, model)
        for i in range(len(relative_error)):
            self.assertLessEqual(relative_error[i], 0.01)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_simple_rnn(self):
        self._test_simple_rnn(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_simple_lstm(self):
        self._test_simple_lstm(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_simple_gru(self):
        self._test_simple_gru(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_simple_rnn(self):
        self._test_simple_rnn(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_simple_lstm(self):
        self._test_simple_lstm(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_simple_gru(self):
        self._test_simple_gru(keras_major_version=2)


class RecurrentLayerTest(unittest.TestCase):
    """
    Base class for recurrent layer tests. Masking param not included here
    """

    def setUp(self):
        self.params_dict = dict(
            input_dims=[[1, 5, 10], [1, 1, 1], [1, 2, 5]],
            output_dim=[1, 5, 10],
            stateful=[False],
            go_backwards=[False, True],
            unroll=[True],
            return_sequences=[False, True],
            activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
        )
        self.base_layer_params = list(itertools.product(*self.params_dict.values()))


class RNNLayer(RecurrentLayerTest):
    """
    Class for testing single RNN layer
    """

    def setUp(self):
        super(RNNLayer, self).setUp()
        self.simple_rnn_params_dict = self.params_dict
        self.rnn_layer_params = list(
            itertools.product(self.simple_rnn_params_dict.values())
        )

    def _test_rnn_layer(self, keras_major_version, limit=None):
        i = 0
        numerical_err_models = []
        shape_err_models = []
        numerical_failiure = 0
        params = list(itertools.product(self.base_layer_params, self.rnn_layer_params))
        np.random.shuffle(params)
        params = [
            param
            for param in params
            if valid_params(dict(zip(self.params_dict.keys(), param[0])))
        ]
        for base_params, rnn_params in params[:limit]:
            base_params = dict(zip(self.params_dict.keys(), base_params))
            rnn_params = dict(zip(self.simple_rnn_params_dict.keys(), rnn_params))
            model = Sequential()
            unroll = base_params["unroll"]
            if base_params["input_dims"][1] == 1 and unroll == True:
                unroll = False
            if keras_major_version == 2:
                model.add(
                    SimpleRNN(
                        base_params["output_dim"],
                        input_shape=base_params["input_dims"][1:],
                        activation=base_params["activation"],
                        return_sequences=base_params["return_sequences"],
                        go_backwards=base_params["go_backwards"],
                        unroll=unroll,
                    )
                )
            else:
                model.add(
                    SimpleRNN(
                        base_params["output_dim"],
                        input_length=base_params["input_dims"][1],
                        input_dim=base_params["input_dims"][2],
                        activation=base_params["activation"],
                        return_sequences=base_params["return_sequences"],
                        go_backwards=base_params["go_backwards"],
                        unroll=unroll,
                    )
                )
            mlkitmodel = get_mlkit_model_from_path(model)
            input_data = generate_input(
                base_params["input_dims"][0],
                base_params["input_dims"][1],
                base_params["input_dims"][2],
            )
            keras_preds = model.predict(input_data).flatten()
            if K.tensorflow_backend._SESSION:
                import tensorflow as tf

                tf.reset_default_graph()
                K.tensorflow_backend._SESSION.close()
                K.tensorflow_backend._SESSION = None
            input_data = np.transpose(input_data, [1, 0, 2])
            if _is_macos() and _macos_version() >= (10, 13):
                coreml_preds = mlkitmodel.predict({"data": input_data})[
                    "output"
                ].flatten()
                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\nbase_params: {}\nkeras_preds.shape: {}\ncoreml_preds.shape: {}".format(
                            base_params, keras_preds.shape, coreml_preds.shape
                        )
                    )
                    shape_err_models.append(base_params)
                    i += 1
                    continue
                try:
                    max_denominator = np.maximum(
                        np.maximum(np.abs(coreml_preds), np.abs(keras_preds)), 1.0
                    )
                    relative_error = (
                        coreml_preds / max_denominator - keras_preds / max_denominator
                    )
                    for i in range(len(relative_error)):
                        self.assertLessEqual(relative_error[i], 0.01)
                except AssertionError:
                    print(
                        "Assertion error:\nbase_params: {}\nkeras_preds: {}\ncoreml_preds: {}".format(
                            base_params, keras_preds, coreml_preds
                        )
                    )
                    numerical_failiure += 1
                    numerical_err_models.append(base_params)
            i += 1

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}\n"
            "Total numerical failiures: {}/{}\n".format(
                numerical_err_models, numerical_failiure, i
            ),
        )

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    @pytest.mark.slow
    def test_kers1_rnn_layer_stress(self):
        self._test_rnn_layer(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_rnn_layer(self):
        self._test_rnn_layer(keras_major_version=1, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_rnn_layer_stress(self):
        self._test_rnn_layer(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_rnn_layer(self):
        self._test_rnn_layer(keras_major_version=2, limit=10)


class LSTMLayer(RecurrentLayerTest):
    """
    Class for testing single RNN layer
    """

    def setUp(self):
        super(LSTMLayer, self).setUp()
        self.lstm_params_dict = dict(
            inner_activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
            bidirectional=[False],
        )
        self.lstm_layer_params = list(
            itertools.product(*self.lstm_params_dict.values())
        )

    def _test_bilstm_layer(self, batched=False):
        if not batched:
            params_dict = dict(
                input_dims=[[1, 5, 10], [1, 2, 5]],
                output_dim=[1, 5, 10],
                activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
                inner_activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
                return_sequences=[True, False],
            )
        else:
            params_dict = dict(
                input_dims=[[3, 5, 10], [6, 2, 5]],
                output_dim=[1, 5, 10],
                activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
                inner_activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
                return_sequences=[True, False],
            )

        params = list(itertools.product(*params_dict.values()))
        ii = 0
        i = 0
        numerical_err_models = []
        shape_err_models = []
        numerical_failiure = 0
        for param in params:
            ii += 1
            # print('-------------- %d / %d ------------------- ' % (ii, len(params)))
            param = dict(zip(params_dict.keys(), param))

            if param["activation"] == "linear":
                keras_act = None
            else:
                keras_act = param["activation"]

            if param["inner_activation"] == "linear":
                keras_inner_act = None
            else:
                keras_inner_act = param["inner_activation"]

            model = Sequential()
            model.add(
                Bidirectional(
                    LSTM(
                        param["output_dim"],
                        activation=keras_act,
                        recurrent_activation=keras_inner_act,
                        return_sequences=param["return_sequences"],
                        go_backwards=False,
                        unroll=False,
                    ),
                    input_shape=(param["input_dims"][1], param["input_dims"][2]),
                )
            )
            mlmodel = get_mlkit_model_from_path(model)

            Batch = param["input_dims"][0]
            Seq = param["input_dims"][1]
            h = param["output_dim"]
            input_size = param["input_dims"][2]

            input_data = generate_input(Batch, Seq, input_size)

            activations_to_test_with_numpy = {"linear", "relu"}
            if (
                param["activation"] in activations_to_test_with_numpy
                or param["inner_activation"] in activations_to_test_with_numpy
            ):
                keras_preds = get_numpy_prediction_bilstm_batched(
                    model, input_data
                )  # (Batch, Seq, h)
            else:
                keras_preds = model.predict(input_data)  # (Batch, Seq, h)

            if _is_macos() and _macos_version() >= (10, 13):
                input_data = np.transpose(input_data, [1, 0, 2])
                input_dict = {}
                input_dict["data"] = input_data
                input_dict["bidirectional_1_h_in"] = np.zeros(
                    (1, Batch, h), dtype=np.float
                )
                input_dict["bidirectional_1_c_in"] = np.zeros(
                    (1, Batch, h), dtype=np.float
                )
                input_dict["bidirectional_1_h_in_rev"] = np.zeros(
                    (1, Batch, h), dtype=np.float
                )
                input_dict["bidirectional_1_c_in_rev"] = np.zeros(
                    (1, Batch, h), dtype=np.float
                )
                coreml_preds = mlmodel.predict(input_dict)[
                    "output"
                ]  # (Seq, Batch, h, .. )
                if param["return_sequences"]:
                    coreml_preds = np.reshape(coreml_preds, [Seq, Batch, 2 * h])
                else:
                    coreml_preds = np.reshape(coreml_preds, [1, Batch, 2 * h])
                    keras_preds = np.expand_dims(keras_preds, axis=1)
                coreml_preds = np.transpose(coreml_preds, [1, 0, 2])

                if K.tensorflow_backend._SESSION:
                    import tensorflow as tf

                    tf.reset_default_graph()
                    K.tensorflow_backend._SESSION.close()
                    K.tensorflow_backend._SESSION = None

                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\n param: {}\n\n keras_preds.shape: {}\n\n coreml_preds.shape: {}".format(
                            param, keras_preds.shape, coreml_preds.shape
                        )
                    )
                    shape_err_models.append(param)
                    i += 1
                    continue
                max_denominator = np.maximum(
                    np.maximum(
                        np.abs(coreml_preds.flatten()), np.abs(keras_preds.flatten())
                    ),
                    1.0,
                )
                relative_error = (
                    coreml_preds.flatten() / max_denominator
                    - keras_preds.flatten() / max_denominator
                )
                max_relative_error = np.amax(relative_error)
                try:
                    self.assertLessEqual(max_relative_error, 0.01)
                except AssertionError:
                    snr, psnr, signal_energy = _compute_SNR(keras_preds, coreml_preds)
                    print("-*" * 80)
                    print("Assertion error. \n param : {} \n".format(param))
                    print(
                        "max error = %.4f, snr = %.1f, psnr = %.1f, energy = %.6f"
                        % (max_relative_error, snr, psnr, signal_energy)
                    )
                    print(
                        "keras preds shape: {}, coreml preds shape = {}".format(
                            str(keras_preds.shape), str(coreml_preds.shape)
                        )
                    )
                    # for b in range(Batch):
                    #     snr, psnr, signal_energy = _compute_SNR(keras_preds[b, :, :], coreml_preds[b, :, :])
                    #     print('snr = %.1f, psnr = %.1f, energy = %.6f' % (snr, psnr, signal_energy))
                    #     print('batch id = {}, keras_preds = \n{} '.format(b, keras_preds[b, :, :]))
                    #     print('batch id = {}, coreml_preds = \n{} '.format(b, coreml_preds[b, :, :]))
                    print("-*" * 80)

                    numerical_failiure += 1
                    numerical_err_models.append(param)
                    continue

            i += 1

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    def _test_batched_lstm_layer(self):
        params_dict = dict(
            input_dims=[[3, 5, 10], [6, 2, 5]],
            output_dim=[1, 5, 10],
            activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
            inner_activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"],
            return_sequences=[True, False],
        )
        params = list(itertools.product(*params_dict.values()))
        ii = 0
        i = 0
        numerical_err_models = []
        shape_err_models = []
        numerical_failiure = 0
        for param in params:
            ii += 1
            # print('-------------- %d / %d ------------------- ' % (ii, len(params)))
            param = dict(zip(params_dict.keys(), param))

            if param["activation"] == "linear":
                keras_act = None
            else:
                keras_act = param["activation"]

            if param["inner_activation"] == "linear":
                keras_inner_act = None
            else:
                keras_inner_act = param["inner_activation"]

            model = Sequential()
            model.add(
                LSTM(
                    param["output_dim"],
                    input_shape=(param["input_dims"][1], param["input_dims"][2]),
                    activation=keras_act,
                    recurrent_activation=keras_inner_act,
                    return_sequences=param["return_sequences"],
                    go_backwards=False,
                    unroll=False,
                )
            )

            mlmodel = get_mlkit_model_from_path(model)

            Batch = param["input_dims"][0]
            Seq = param["input_dims"][1]
            h = param["output_dim"]
            input_size = param["input_dims"][2]

            input_data = generate_input(Batch, Seq, input_size)

            keras_preds = model.predict(input_data)  # (Batch, Seq, h)

            if _is_macos() and _macos_version() >= (10, 13):
                input_data = np.transpose(input_data, [1, 0, 2])
                input_dict = {}
                input_dict["data"] = input_data
                input_dict["lstm_1_h_in"] = np.zeros((1, Batch, h), dtype=np.float)
                input_dict["lstm_1_c_in"] = np.zeros((1, Batch, h), dtype=np.float)
                coreml_preds = mlmodel.predict(input_dict)[
                    "output"
                ]  # (Seq, Batch, h, .. )
                if param["return_sequences"]:
                    coreml_preds = np.reshape(coreml_preds, [Seq, Batch, h])
                else:
                    coreml_preds = np.reshape(coreml_preds, [1, Batch, h])
                    keras_preds = np.expand_dims(keras_preds, axis=1)
                coreml_preds = np.transpose(coreml_preds, [1, 0, 2])

                if K.tensorflow_backend._SESSION:
                    import tensorflow as tf

                    tf.reset_default_graph()
                    K.tensorflow_backend._SESSION.close()
                    K.tensorflow_backend._SESSION = None

                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\n param: {}\n\n keras_preds.shape: {}\n\n coreml_preds.shape: {}".format(
                            param, keras_preds.shape, coreml_preds.shape
                        )
                    )
                    shape_err_models.append(param)
                    i += 1
                    continue
                try:
                    max_denominator = np.maximum(
                        np.maximum(
                            np.abs(coreml_preds.flatten()),
                            np.abs(keras_preds.flatten()),
                        ),
                        1.0,
                    )
                    relative_error = (
                        coreml_preds.flatten() / max_denominator
                        - keras_preds.flatten() / max_denominator
                    )
                    max_relative_error = np.amax(relative_error)
                    self.assertLessEqual(max_relative_error, 0.01)
                except AssertionError:
                    snr, psnr, signal_energy = _compute_SNR(keras_preds, coreml_preds)
                    print("-*" * 80)
                    print("Assertion error. \n param : {} \n".format(param))
                    print(
                        "max error = %.4f, snr = %.1f, psnr = %.1f, energy = %.6f"
                        % (max_relative_error, snr, psnr, signal_energy)
                    )
                    print(
                        "keras preds shape: {}, coreml preds shape = {}".format(
                            str(keras_preds.shape), str(coreml_preds.shape)
                        )
                    )
                    # for b in range(Batch):
                    #     snr, psnr, signal_energy = _compute_SNR(keras_preds[b, :, :], coreml_preds[b, :, :])
                    #     print('snr = %.1f, psnr = %.1f, energy = %.6f' % (snr, psnr, signal_energy))
                    #     print('batch id = {}, keras_preds = \n{} '.format(b, keras_preds[b, :, :]))
                    #     print('batch id = {}, coreml_preds = \n{} '.format(b, coreml_preds[b, :, :]))
                    print("-*" * 80)

                    numerical_failiure += 1
                    numerical_err_models.append(param)
                    continue

            i += 1

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    def _test_lstm_layer(self, keras_major_version, limit=None):
        params_keys = list(self.params_dict.keys())
        numerical_err_models = []
        shape_err_models = []
        numerical_failiure = 0
        params = list(itertools.product(self.base_layer_params, self.lstm_layer_params))
        np.random.shuffle(params)

        params = [
            param
            for param in params
            if valid_params(dict(zip(self.params_dict.keys(), param[0])))
        ]
        ctr = 0
        for base_params, lstm_params in params[:limit]:
            ctr += 1
            # print('--------------- Testing %d/%d ---------------' %(ctr, len(params)))
            base_params = dict(zip(self.params_dict.keys(), base_params))
            lstm_params = dict(zip(self.lstm_params_dict.keys(), lstm_params))
            model = Sequential()
            unroll = base_params["unroll"]
            if base_params["input_dims"][1] == 1 and unroll == True:
                unroll = False
            if lstm_params["bidirectional"] is True:
                if keras_major_version == 2:
                    model.add(
                        Bidirectional(
                            LSTM(
                                base_params["output_dim"],
                                activation=base_params["activation"],
                                recurrent_activation=lstm_params["inner_activation"],
                                return_sequences=base_params["return_sequences"],
                                go_backwards=False,
                                unroll=unroll,
                            ),
                            input_shape=(
                                base_params["input_dims"][1],
                                base_params["input_dims"][2],
                            ),
                        )
                    )
                else:
                    model.add(
                        Bidirectional(
                            LSTM(
                                base_params["output_dim"],
                                activation=base_params["activation"],
                                inner_activation=lstm_params["inner_activation"],
                                return_sequences=base_params["return_sequences"],
                                go_backwards=False,
                                unroll=unroll,
                            ),
                            input_shape=(
                                base_params["input_dims"][1],
                                base_params["input_dims"][2],
                            ),
                        )
                    )
            else:
                if keras_major_version == 2:
                    model.add(
                        LSTM(
                            base_params["output_dim"],
                            input_shape=(
                                base_params["input_dims"][1],
                                base_params["input_dims"][2],
                            ),
                            activation=base_params["activation"],
                            recurrent_activation=lstm_params["inner_activation"],
                            return_sequences=base_params["return_sequences"],
                            go_backwards=base_params["go_backwards"],
                            unroll=unroll,
                        )
                    )
                else:
                    model.add(
                        LSTM(
                            base_params["output_dim"],
                            input_shape=(
                                base_params["input_dims"][1],
                                base_params["input_dims"][2],
                            ),
                            activation=base_params["activation"],
                            inner_activation=lstm_params["inner_activation"],
                            return_sequences=base_params["return_sequences"],
                            go_backwards=base_params["go_backwards"],
                            unroll=unroll,
                        )
                    )
            mlkitmodel = get_mlkit_model_from_path(model)
            input_data = generate_input(
                base_params["input_dims"][0],
                base_params["input_dims"][1],
                base_params["input_dims"][2],
            )

            activations_to_test_with_numpy = {"linear", "relu"}
            if (
                base_params["activation"] in activations_to_test_with_numpy
                or lstm_params["inner_activation"] in activations_to_test_with_numpy
            ):
                if lstm_params["bidirectional"]:
                    keras_preds = get_numpy_prediction_bilstm(
                        model, input_data
                    ).flatten()
                else:
                    keras_preds = get_numpy_prediction_unilstm(
                        model, input_data
                    ).flatten()
            else:
                keras_preds = model.predict(input_data).flatten()

            if _is_macos() and _macos_version() >= (10, 13):
                input_data = np.transpose(input_data, [1, 0, 2])
                coreml_preds = mlkitmodel.predict({"data": input_data})[
                    "output"
                ].flatten()

                if K.tensorflow_backend._SESSION:
                    import tensorflow as tf

                    tf.reset_default_graph()
                    K.tensorflow_backend._SESSION.close()
                    K.tensorflow_backend._SESSION = None

                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\n base_params: {}\n\n lstm_params: {}\n\n keras_preds.shape: {}\n\n coreml_preds.shape: {}".format(
                            base_params,
                            lstm_params,
                            keras_preds.shape,
                            coreml_preds.shape,
                        )
                    )
                    shape_err_models.append(base_params)
                    continue

                max_denominator = np.maximum(
                    np.maximum(np.abs(coreml_preds), np.abs(keras_preds)), 1.0
                )
                try:
                    relative_error = (
                        coreml_preds / max_denominator - keras_preds / max_denominator
                    )
                    for i in range(len(relative_error)):
                        self.assertLessEqual(relative_error[i], 0.01)
                except AssertionError:
                    print(
                        "Assertion error:\n base_params: {}\n lstm_params: {}\n\n keras_preds: {}\n\n coreml_preds: {}\n\n\n keras_preds: {}\n\n\n coreml_preds: {}\n".format(
                            base_params,
                            lstm_params,
                            keras_preds / max_denominator,
                            coreml_preds / max_denominator,
                            keras_preds,
                            coreml_preds,
                        )
                    )
                    numerical_failiure += 1
                    numerical_err_models.append(base_params)

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    @pytest.mark.slow
    def test_keras_lstm_layer_stress(self):
        self._test_lstm_layer(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras_lstm_layer(self):
        self._test_lstm_layer(keras_major_version=1, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_lstm_layer_stress(self):
        self._test_lstm_layer(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_lstm_layer(self):
        self._test_lstm_layer(keras_major_version=2, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_bilstm_layer(self):
        self._test_bilstm_layer()

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_bilstm_layer_batched(self):
        self._test_bilstm_layer(batched=True)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_lstm_layer_batched(self):
        self._test_batched_lstm_layer()


class GRULayer(RecurrentLayerTest):
    """
    Class for testing GRU layer
    """

    def setUp(self):
        super(GRULayer, self).setUp()
        self.gru_params_dict = dict(
            inner_activation=["tanh", "linear", "sigmoid", "hard_sigmoid", "relu"]
        )
        self.gru_layer_params = list(itertools.product(*self.gru_params_dict.values()))

    def _test_gru_layer(self, keras_major_version, limit=None):
        i = 0
        numerical_err_models = []
        shape_err_models = []
        numerical_failiure = 0
        params = list(itertools.product(self.base_layer_params, self.gru_layer_params))
        np.random.shuffle(params)
        params = [
            param
            for param in params
            if valid_params(dict(zip(self.params_dict.keys(), param[0])))
        ]
        for base_params, gru_params in params[:limit]:
            base_params = dict(zip(self.params_dict.keys(), base_params))
            gru_params = dict(zip(self.gru_params_dict.keys(), gru_params))
            model = Sequential()
            unroll = base_params["unroll"]
            if base_params["input_dims"][1] == 1 and unroll == True:
                unroll = False
            if keras_major_version == 2:
                model.add(
                    GRU(
                        base_params["output_dim"],
                        input_shape=(
                            base_params["input_dims"][1],
                            base_params["input_dims"][2],
                        ),
                        activation=base_params["activation"],
                        recurrent_activation=gru_params["inner_activation"],
                        return_sequences=base_params["return_sequences"],
                        go_backwards=base_params["go_backwards"],
                        unroll=unroll,
                    )
                )
            else:
                model.add(
                    GRU(
                        base_params["output_dim"],
                        input_length=base_params["input_dims"][1],
                        input_dim=base_params["input_dims"][2],
                        activation=base_params["activation"],
                        inner_activation=gru_params["inner_activation"],
                        return_sequences=base_params["return_sequences"],
                        go_backwards=base_params["go_backwards"],
                        unroll=unroll,
                    )
                )
            model.set_weights([np.random.rand(*w.shape) for w in model.get_weights()])
            mlkitmodel = get_mlkit_model_from_path(model)
            input_data = generate_input(
                base_params["input_dims"][0],
                base_params["input_dims"][1],
                base_params["input_dims"][2],
            )

            activations_to_test_with_numpy = {"linear", "relu"}
            if (
                base_params["activation"] in activations_to_test_with_numpy
                or gru_params["inner_activation"] in activations_to_test_with_numpy
            ):
                keras_preds = get_numpy_prediction_gru(model, input_data).flatten()
            else:
                keras_preds = model.predict(input_data).flatten()

            if _is_macos() and _macos_version() >= (10, 13):
                input_data = np.transpose(input_data, [1, 0, 2])
                coreml_preds = mlkitmodel.predict({"data": input_data})[
                    "output"
                ].flatten()
                if K.tensorflow_backend._SESSION:
                    import tensorflow as tf

                    tf.reset_default_graph()
                    K.tensorflow_backend._SESSION.close()
                    K.tensorflow_backend._SESSION = None
                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\nbase_params: {}\n gru_params: {}\nkeras_preds.shape: {}\ncoreml_preds.shape: {}".format(
                            base_params,
                            gru_params,
                            keras_preds.shape,
                            coreml_preds.shape,
                        )
                    )
                    shape_err_models.append(base_params)
                    i += 1
                    continue

                max_denominator = np.maximum(
                    np.maximum(np.abs(coreml_preds), np.abs(keras_preds)), 1.0
                )
                try:
                    relative_error = (
                        coreml_preds / max_denominator - keras_preds / max_denominator
                    )
                    for i in range(len(relative_error)):
                        self.assertLessEqual(relative_error[i], 0.01)
                except AssertionError:
                    print(
                        "===============Assertion error:\n base_params: {}\n gru_params: {}\n\n keras_preds: {}\n\n coreml_preds: {}\n\n\n keras_preds: {}\n\n\n coreml_preds: {}\n".format(
                            base_params,
                            gru_params,
                            keras_preds / max_denominator,
                            coreml_preds / max_denominator,
                            keras_preds,
                            coreml_preds,
                        )
                    )
                    numerical_failiure += 1
                    numerical_err_models.append(base_params)
            i += 1

        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    @pytest.mark.slow
    def test_keras1_test_gru_layer_stress(self):
        self._test_gru_layer(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_test_gru_layer(self):
        self._test_gru_layer(keras_major_version=1, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_test_gru_layer_stress(self):
        self._test_gru_layer(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_test_gru_layer(self):
        self._test_gru_layer(keras_major_version=2, limit=10)


class LSTMStacked(unittest.TestCase):
    """
    Class for testing LSTMStacked
    """

    def setUp(self):
        self.params_dict = dict(
            input_dims=[[1, 1, 1], [1, 2, 5], [1, 5, 10]],
            output_dim=[1, 5, 10, 20],
            stateful=[False],
            go_backwards=[False],
            unroll=[True],
            return_sequences=[True],
            top_return_sequences=[True, False],
            activation=["tanh", "sigmoid", "hard_sigmoid"],
            number_of_layers=[1, 2, 3],
        )
        self.base_layer_params = list(itertools.product(*self.params_dict.values()))

    def _test_lstm_stacked(self, keras_major_version, limit=None):
        numerical_err_models = []
        shape_err_models = []
        numerical_failiure = 0
        params = copy(self.base_layer_params)
        np.random.shuffle(params)
        i = 0
        params = [
            param
            for param in params
            if valid_params(dict(zip(self.params_dict.keys(), param)))
        ]
        for base_params in params[:limit]:
            base_params = dict(zip(self.params_dict.keys(), base_params))
            model = Sequential()
            unroll = base_params["unroll"]
            if base_params["input_dims"][1] == 1 and unroll == True:
                unroll = False
            settings = dict(
                activation=base_params["activation"],
                return_sequences=True,
                go_backwards=base_params["go_backwards"],
                unroll=unroll,
            )
            if keras_major_version == 2:
                model.add(
                    LSTM(
                        base_params["output_dim"],
                        input_shape=base_params["input_dims"][1:],
                        recurrent_activation="sigmoid",
                        **settings
                    )
                )
                for idx in range(0, base_params["number_of_layers"]):
                    model.add(
                        LSTM(
                            base_params["output_dim"],
                            input_shape=(
                                base_params["input_dims"][1],
                                base_params["output_dim"],
                            ),
                            return_sequences=True,
                            activation="tanh",
                            recurrent_activation="sigmoid",
                        )
                    )
                model.add(
                    LSTM(
                        10,
                        input_shape=(
                            base_params["input_dims"][1],
                            base_params["output_dim"],
                        ),
                        return_sequences=base_params["top_return_sequences"],
                        activation="sigmoid",
                    )
                )

            else:
                model.add(
                    LSTM(
                        output_dim=base_params["output_dim"],
                        input_length=base_params["input_dims"][1],
                        input_dim=base_params["input_dims"][2],
                        inner_activation="sigmoid",
                        **settings
                    )
                )
                for idx in range(0, base_params["number_of_layers"]):
                    model.add(
                        LSTM(
                            output_dim=base_params["output_dim"],
                            return_sequences=True,
                            activation="tanh",
                            inner_activation="sigmoid",
                        )
                    )
                model.add(
                    LSTM(
                        output_dim=10,
                        return_sequences=base_params["top_return_sequences"],
                        activation="sigmoid",
                    )
                )
            mlkitmodel = get_mlkit_model_from_path(model)
            input_data = generate_input(
                base_params["input_dims"][0],
                base_params["input_dims"][1],
                base_params["input_dims"][2],
            )
            if _is_macos() and _macos_version() >= (10, 13):
                keras_preds = model.predict(input_data).flatten()
                input_data = np.transpose(input_data, [1, 0, 2])
                coreml_preds = mlkitmodel.predict({"data": input_data})[
                    "output"
                ].flatten()
                import tensorflow as tf

                tf.reset_default_graph()
                K.tensorflow_backend._SESSION.close()
                K.tensorflow_backend._SESSION = None
                try:
                    self.assertEquals(coreml_preds.shape, keras_preds.shape)
                except AssertionError:
                    print(
                        "Shape error:\nbase_params: {}\nkeras_preds.shape: {}\ncoreml_preds.shape: {}".format(
                            base_params, keras_preds.shape, coreml_preds.shape
                        )
                    )
                    shape_err_models.append(base_params)
                    i += 1
                    continue
                try:
                    max_denominator = np.maximum(
                        np.maximum(np.abs(coreml_preds), np.abs(keras_preds)), 1.0
                    )
                    relative_error = (
                        coreml_preds / max_denominator - keras_preds / max_denominator
                    )
                    for i in range(len(relative_error)):
                        self.assertLessEqual(relative_error[i], 0.01)
                except AssertionError:
                    print(
                        "Assertion error:\nbase_params: {}\nkeras_preds: {}\ncoreml_preds: {}".format(
                            base_params, keras_preds, coreml_preds
                        )
                    )
                    numerical_failiure += 1
                    numerical_err_models.append(base_params)
            i += 1
        self.assertEquals(
            shape_err_models, [], msg="Shape error models {}".format(shape_err_models)
        )
        self.assertEquals(
            numerical_err_models,
            [],
            msg="Numerical error models {}".format(numerical_err_models),
        )

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    @pytest.mark.slow
    def test_keras1_lstm_stacked_stress(self):
        self._test_lstm_stacked(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_lstm_stacked(self):
        self._test_lstm_stacked(keras_major_version=1, limit=10)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    @pytest.mark.slow
    def test_keras2_lstm_stacked_stress(self):
        self._test_lstm_stacked(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_lstm_stacked(self):
        self._test_lstm_stacked(keras_major_version=2, limit=10)


class DifferentIOModelsTypes(unittest.TestCase):
    """
    Class for testing different I/O combinations for LSTMS
    """

    def _test_one_to_many(self, keras_major_version):
        params = (
            dict(
                input_dims=[1, 10],
                activation="tanh",
                return_sequences=False,
                output_dim=3,
            ),
        )
        number_of_times = 4
        model = Sequential()
        model.add(RepeatVector(number_of_times, input_shape=(10,)))

        if keras_major_version == 2:
            model.add(
                LSTM(
                    params[0]["output_dim"],
                    input_shape=params[0]["input_dims"],
                    activation=params[0]["activation"],
                    recurrent_activation="sigmoid",
                    return_sequences=True,
                )
            )
        else:
            model.add(
                LSTM(
                    output_dim=params[0]["output_dim"],
                    activation=params[0]["activation"],
                    inner_activation="sigmoid",
                    return_sequences=True,
                )
            )
        relative_error, keras_preds, coreml_preds = simple_model_eval(params, model)
        # print relative_error, '\n', keras_preds, '\n', coreml_preds, '\n'
        for i in range(len(relative_error)):
            self.assertLessEqual(relative_error[i], 0.01)

    def _test_many_to_one(self, keras_major_version):
        params = (
            dict(
                input_dims=[1, 10, 5],
                go_backwards=False,
                activation="tanh",  # fails with hard_sigmoid
                stateful=False,
                unroll=False,
                return_sequences=False,
                output_dim=1,
            ),
        )
        model = Sequential()
        if keras_major_version == 2:
            model.add(
                LSTM(
                    params[0]["output_dim"],
                    input_shape=params[0]["input_dims"][1:],
                    activation=params[0]["activation"],
                    recurrent_activation="sigmoid",
                )
            )
        else:
            model.add(
                LSTM(
                    output_dim=params[0]["output_dim"],
                    input_shape=params[0]["input_dims"][1:],
                    activation=params[0]["activation"],
                    inner_activation="sigmoid",
                )
            )
        relative_error, keras_preds, coreml_preds = simple_model_eval(params, model)
        # print relative_error, '\n', keras_preds, '\n', coreml_preds, '\n'
        for i in range(len(relative_error)):
            self.assertLessEqual(relative_error[i], 0.01)

    def _test_many_to_many(self, keras_major_version):
        params = (
            dict(
                input_dims=[1, 10, 5],
                go_backwards=False,
                activation="tanh",  # fails with hard_sigmoid
                stateful=False,
                unroll=False,
                return_sequences=True,
                output_dim=1,
            ),
        )
        model = Sequential()
        if keras_major_version == 2:
            model.add(
                LSTM(
                    params[0]["output_dim"],
                    input_shape=params[0]["input_dims"][1:],
                    activation=params[0]["activation"],
                    recurrent_activation="sigmoid",
                    return_sequences=True,
                )
            )
        else:
            model.add(
                LSTM(
                    output_dim=params[0]["output_dim"],
                    input_shape=params[0]["input_dims"][1:],
                    activation=params[0]["activation"],
                    inner_activation="sigmoid",
                    return_sequences=True,
                )
            )
        relative_error, keras_preds, coreml_preds = simple_model_eval(params, model)
        # print relative_error, '\n', keras_preds, '\n', coreml_preds, '\n'
        for i in range(len(relative_error)):
            self.assertLessEqual(relative_error[i], 0.01)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_test_one_to_many(self):
        self._test_one_to_many(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_test_many_to_one(self):
        self._test_many_to_one(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS_TF, "Missing keras 1. Skipping test.")
    @pytest.mark.keras1
    def test_keras1_many_to_many(self):
        self._test_many_to_many(keras_major_version=1)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_test_one_to_many(self):
        self._test_one_to_many(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_test_many_to_one(self):
        self._test_many_to_one(keras_major_version=2)

    @unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
    @pytest.mark.keras2
    def test_keras2_many_to_many(self):
        self._test_many_to_many(keras_major_version=2)


@unittest.skipIf(not _HAS_KERAS2_TF, "Missing keras 2. Skipping test.")
@pytest.mark.keras2
class InitialStateRecurrentModels(unittest.TestCase):
    """
    This test class sets initial states to the recurrent nodes and then test
    """

    @unittest.skip("failing - TODO re-enable when it passes consistently")
    def test_initial_state_GRU(self):
        data = np.random.rand(1, 1, 2)

        model = keras.models.Sequential()
        model.add(
            keras.layers.GRU(
                5, input_shape=(1, 2), batch_input_shape=[1, 1, 2], stateful=True
            )
        )
        model.get_layer(index=1).reset_states()

        coreml_model = keras_converter.convert(
            model=model, input_names="data", output_names="output"
        )
        if _is_macos() and _macos_version() >= (10, 13):
            keras_output_1 = model.predict(data)
            coreml_full_output_1 = coreml_model.predict({"data": data})
            coreml_output_1 = coreml_full_output_1["output"]
            coreml_output_1 = np.expand_dims(coreml_output_1, 1)

            np.testing.assert_array_almost_equal(coreml_output_1.T, keras_output_1)

        hidden_state = np.random.rand(1, 5)
        model.get_layer(index=1).reset_states(hidden_state)
        coreml_model = keras_converter.convert(
            model=model, input_names="data", output_names="output"
        )
        spec = coreml_model.get_spec()
        if _is_macos() and _macos_version() >= (10, 13):
            keras_output_2 = model.predict(data)
            coreml_full_output_2 = coreml_model.predict(
                {"data": data, spec.description.input[1].name: hidden_state[0]}
            )
            coreml_output_2 = coreml_full_output_2["output"]
            coreml_output_2 = np.expand_dims(coreml_output_2, 1)
            np.testing.assert_array_almost_equal(coreml_output_2.T, keras_output_2)

    def test_initial_state_SimpleRNN(self):
        data = np.random.rand(1, 1, 2)
        model = keras.models.Sequential()
        model.add(
            keras.layers.SimpleRNN(
                5, input_shape=(1, 2), batch_input_shape=[1, 1, 2], stateful=True
            )
        )
        model.get_layer(index=1).reset_states()
        coreml_model = keras_converter.convert(
            model=model, input_names="data", output_names="output"
        )
        if _is_macos() and _macos_version() >= (10, 13):
            keras_output_1 = model.predict(data)
            coreml_full_output_1 = coreml_model.predict({"data": data})
            coreml_output_1 = coreml_full_output_1["output"]
            coreml_output_1 = np.expand_dims(coreml_output_1, 1)
            np.testing.assert_array_almost_equal(coreml_output_1.T, keras_output_1)

        hidden_state = np.random.rand(1, 5)
        model.get_layer(index=1).reset_states(hidden_state)
        coreml_model = keras_converter.convert(
            model=model, input_names="data", output_names="output"
        )
        spec = coreml_model.get_spec()
        if _is_macos() and _macos_version() >= (10, 13):
            keras_output_2 = model.predict(data)
            coreml_full_output_2 = coreml_model.predict(
                {"data": data, spec.description.input[1].name: hidden_state[0]}
            )
            coreml_output_2 = coreml_full_output_2["output"]
            coreml_output_2 = np.expand_dims(coreml_output_2, 1)
            np.testing.assert_array_almost_equal(coreml_output_2.T, keras_output_2)

    def test_initial_state_LSTM(self):
        data = np.random.rand(1, 1, 2)

        model = keras.models.Sequential()
        model.add(
            keras.layers.LSTM(
                5, input_shape=(1, 2), batch_input_shape=[1, 1, 2], stateful=True
            )
        )
        model.get_layer(index=1).reset_states()

        if _is_macos() and _macos_version() >= (10, 13):
            coreml_model = keras_converter.convert(
                model=model, input_names="data", output_names="output"
            )

            keras_output_1 = model.predict(data)
            coreml_full_output_1 = coreml_model.predict({"data": data})
            coreml_output_1 = coreml_full_output_1["output"]
            coreml_output_1 = np.expand_dims(coreml_output_1, 1)

            np.testing.assert_array_almost_equal(coreml_output_1.T, keras_output_1)

        hidden_state = (np.random.rand(1, 5), np.random.rand(1, 5))
        model.get_layer(index=1).reset_states(hidden_state)

        coreml_model = keras_converter.convert(
            model=model, input_names="data", output_names="output"
        )
        spec = coreml_model.get_spec()

        if _is_macos() and _macos_version() >= (10, 13):
            keras_output_2 = model.predict(data)
            coreml_full_output_2 = coreml_model.predict(
                {
                    "data": data,
                    spec.description.input[1].name: hidden_state[0][0],
                    spec.description.input[2].name: hidden_state[1][0],
                }
            )
            coreml_output_2 = coreml_full_output_2["output"]
            coreml_output_2 = np.expand_dims(coreml_output_2, 1)

            np.testing.assert_array_almost_equal(coreml_output_2.T, keras_output_2)


if __name__ == "__main__":
    # unittest.main()
    ## To run a specific test:
    suite = unittest.TestSuite()
    suite.addTest(LSTMLayer("test_keras2_bilstm_layer"))
    unittest.TextTestRunner().run(suite)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SoundAnalysisPreprocessing.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='SoundAnalysisPreprocessing.proto',
  package='CoreML.Specification.CoreMLModels',
  syntax='proto3',
  serialized_pb=_b('\n SoundAnalysisPreprocessing.proto\x12!CoreML.Specification.CoreMLModels\"\xa0\x01\n\x1aSoundAnalysisPreprocessing\x12V\n\x06vggish\x18\x14 \x01(\x0b\x32\x44.CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.VggishH\x00\x1a\x08\n\x06VggishB \n\x1eSoundAnalysisPreprocessingTypeB\x02H\x03\x62\x06proto3')
)




_SOUNDANALYSISPREPROCESSING_VGGISH = _descriptor.Descriptor(
  name='Vggish',
  full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.Vggish',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=190,
  serialized_end=198,
)

_SOUNDANALYSISPREPROCESSING = _descriptor.Descriptor(
  name='SoundAnalysisPreprocessing',
  full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vggish', full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.vggish', index=0,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_SOUNDANALYSISPREPROCESSING_VGGISH, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='SoundAnalysisPreprocessingType', full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.SoundAnalysisPreprocessingType',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=72,
  serialized_end=232,
)

_SOUNDANALYSISPREPROCESSING_VGGISH.containing_type = _SOUNDANALYSISPREPROCESSING
_SOUNDANALYSISPREPROCESSING.fields_by_name['vggish'].message_type = _SOUNDANALYSISPREPROCESSING_VGGISH
_SOUNDANALYSISPREPROCESSING.oneofs_by_name['SoundAnalysisPreprocessingType'].fields.append(
  _SOUNDANALYSISPREPROCESSING.fields_by_name['vggish'])
_SOUNDANALYSISPREPROCESSING.fields_by_name['vggish'].containing_oneof = _SOUNDANALYSISPREPROCESSING.oneofs_by_name['SoundAnalysisPreprocessingType']
DESCRIPTOR.message_types_by_name['SoundAnalysisPreprocessing'] = _SOUNDANALYSISPREPROCESSING
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SoundAnalysisPreprocessing = _reflection.GeneratedProtocolMessageType('SoundAnalysisPreprocessing', (_message.Message,), dict(

  Vggish = _reflection.GeneratedProtocolMessageType('Vggish', (_message.Message,), dict(
    DESCRIPTOR = _SOUNDANALYSISPREPROCESSING_VGGISH,
    __module__ = 'SoundAnalysisPreprocessing_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.Vggish)
    ))
  ,
  DESCRIPTOR = _SOUNDANALYSISPREPROCESSING,
  __module__ = 'SoundAnalysisPreprocessing_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing)
  ))
_sym_db.RegisterMessage(SoundAnalysisPreprocessing)
_sym_db.RegisterMessage(SoundAnalysisPreprocessing.Vggish)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest

import pandas as pd
import itertools
import pytest

from coremltools._deps import _HAS_SKLEARN, _HAS_XGBOOST
from coremltools.models.utils import evaluate_regressor, _macos_version, _is_macos

if _HAS_XGBOOST:
    import xgboost
    from coremltools.converters import xgboost as xgb_converter

if _HAS_SKLEARN:
    from sklearn.datasets import load_boston
    from sklearn.ensemble import GradientBoostingRegressor
    from coremltools.converters import sklearn as skl_converter
    from sklearn.tree import DecisionTreeRegressor

# Skipping tests due to <rdar://problem/66545691> xgboost/test_boosted_trees_*.py failing on CI
import pytest
pytestmark = pytest.mark.skip("Skipping for now. See 66545691 for details.")
#

@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class GradientBoostingRegressorBostonHousingScikitNumericTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Load data and train model
        scikit_data = load_boston()
        self.scikit_data = scikit_data
        self.X = scikit_data["data"]
        self.target = scikit_data["target"]
        self.feature_names = scikit_data.feature_names
        self.output_name = "target"

    def _check_metrics(self, metrics, params={}):
        self.assertAlmostEquals(
            metrics["rmse"],
            0,
            delta=1e-5,
            msg="Failed case %s. Results %s" % (params, metrics),
        )
        self.assertAlmostEquals(
            metrics["max_error"],
            0,
            delta=1e-5,
            msg="Failed case %s. Results %s" % (params, metrics),
        )

    def _train_convert_evaluate_assert(self, **scikit_params):
        scikit_model = GradientBoostingRegressor(random_state=1, **scikit_params)
        scikit_model.fit(self.X, self.target)

        # Convert the model
        spec = skl_converter.convert(scikit_model, self.feature_names, self.output_name)

        if _is_macos() and _macos_version() >= (10, 13):
            # Get predictions
            df = pd.DataFrame(self.X, columns=self.feature_names)
            df["prediction"] = scikit_model.predict(self.X)

            # Evaluate it
            metrics = evaluate_regressor(spec, df, "target", verbose=False)
            self._check_metrics(metrics, scikit_params)

    def test_boston_housing_simple_regression(self):
        self._train_convert_evaluate_assert()

    @pytest.mark.slow
    def test_boston_housing_parameter_stress_test(self):

        options = dict(
            max_depth=[1, 10, None],
            min_samples_split=[2, 0.5],
            min_samples_leaf=[1, 5],
            min_weight_fraction_leaf=[0.0, 0.5],
            max_features=[None, 1],
            max_leaf_nodes=[None, 20],
        )

        # Make a cartesian product of all options
        product = itertools.product(*options.values())
        args = [dict(zip(options.keys(), p)) for p in product]

        print("Testing a total of %s cases. This could take a while" % len(args))
        for it, arg in enumerate(args):
            self._train_convert_evaluate_assert(**arg)


@unittest.skipIf(not _HAS_XGBOOST, "Missing xgboost. Skipping")
@unittest.skipIf(not _HAS_SKLEARN, "Missing scikit-learn. Skipping tests.")
class XgboostBoosterBostonHousingNumericTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        if not _HAS_XGBOOST:
            return
        if not _HAS_SKLEARN:
            return

        # Load data and train model
        scikit_data = load_boston()
        self.X = scikit_data.data.astype("f").astype("d")
        self.dtrain = xgboost.DMatrix(
            scikit_data.data,
            label=scikit_data.target,
            feature_names=scikit_data.feature_names,
        )
        self.feature_names = scikit_data.feature_names
        self.output_name = "target"

    def _check_metrics(self, metrics, allowed_error={}, params={}):
        """
        Check the metrics
        """
        self.assertAlmostEquals(
            metrics["rmse"],
            allowed_error.get("rmse", 0),
            delta=1e-2,
            msg="Failed case %s. Results %s" % (params, metrics),
        )
        self.assertAlmostEquals(
            metrics["max_error"],
            allowed_error.get("max_error", 0),
            delta=1e-2,
            msg="Failed case %s. Results %s" % (params, metrics),
        )

    def _train_convert_evaluate_assert(self, bt_params={}, allowed_error={}, **params):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        # Train a model
        xgb_model = xgboost.train(bt_params, self.dtrain, **params)

        # Convert the model
        spec = xgb_converter.convert(
            xgb_model, self.feature_names, self.output_name, force_32bit_float=False
        )

        if _is_macos() and _macos_version() >= (10, 13):
            # Get predictions
            df = pd.DataFrame(self.X, columns=self.feature_names)
            df["prediction"] = xgb_model.predict(self.dtrain)

            # Evaluate it
            metrics = evaluate_regressor(spec, df, target="target", verbose=False)
            self._check_metrics(metrics, allowed_error, bt_params)

    def test_boston_housing_simple_decision_tree_regression(self):
        self._train_convert_evaluate_assert(num_boost_round=1)

    def test_boston_housing_simple_boosted_tree_regression(self):
        self._train_convert_evaluate_assert(num_boost_round=10)

    def test_boston_housing_simple_random_forest_regression(self):
        self._train_convert_evaluate_assert(bt_params={"subsample": 0.5},
                                            allowed_error={"rmse": 0.004, "max_error": 0.09})

    def test_boston_housing_float_double_corner_case(self):
        self._train_convert_evaluate_assert(
            {
                "colsample_bytree": 1,
                "colsample_bylevel": 1,
                "scale_pos_weight": 1,
                "learning_rate": 0.5,
                "max_delta_step": 0,
                "min_child_weight": 1,
                "n_estimators": 1,
                "subsample": 0.5,
                "objective": "reg:linear",
                "max_depth": 5,
            },
            num_boost_round=2,
        )

    @pytest.mark.slow
    def test_boston_housing_parameter_stress_test(self):

        options = dict(
            max_depth=[1, 5],
            learning_rate=[0.1, 0.5],
            n_estimators=[1, 10],
            min_child_weight=[1, 2],
            max_delta_step=[0, 0.1],
            colsample_bytree=[1, 0.5],
            colsample_bylevel=[1, 0.5],
            scale_pos_weight=[1],
            objective=["reg:linear"],
        )

        # Make a cartesian product of all options
        product = itertools.product(*options.values())
        args = [dict(zip(options.keys(), p)) for p in product]

        print("Testing a total of %s cases. This could take a while" % len(args))
        for it, arg in enumerate(args):
            self._train_convert_evaluate_assert(arg)


@unittest.skipIf(not _HAS_XGBOOST, "Missing xgboost. Skipping")
@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class XGboostRegressorBostonHousingNumericTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """

        # Load data and train model
        scikit_data = load_boston()

        self.X = scikit_data.data
        self.scikit_data = self.X
        self.target = scikit_data.target
        self.feature_names = scikit_data.feature_names
        self.output_name = "target"

    def _check_metrics(self, metrics, params={}, allowed_error={}):
        self.assertAlmostEquals(
            metrics["rmse"],
            allowed_error.get("rmse", 0),
            delta=1e-2,
            msg="Failed case %s. Results %s" % (params, metrics),
        )
        self.assertAlmostEquals(
            metrics["max_error"],
            allowed_error.get("max_error", 0),
            delta=1e-2,
            msg="Failed case %s. Results %s" % (params, metrics),
        )

    def _train_convert_evaluate_assert(self, bt_params={}, allowed_error={}, **params):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        # Train a model
        xgb_model = xgboost.XGBRegressor(**params)
        xgb_model.fit(self.X, self.target)

        # Convert the model (feature_names can't be given because of XGboost)
        spec = xgb_converter.convert(
            xgb_model, self.feature_names, self.output_name, force_32bit_float=False
        )

        if _is_macos() and _macos_version() >= (10, 13):
            # Get predictions
            df = pd.DataFrame(self.X, columns=self.feature_names)
            df["prediction"] = xgb_model.predict(self.X)

            # Evaluate it
            metrics = evaluate_regressor(spec, df, target="target", verbose=False)
            self._check_metrics(metrics, bt_params, allowed_error)

    def test_boston_housing_simple_boosted_tree_regression(self):
        self._train_convert_evaluate_assert()

    def test_boston_housing_simple_random_forest_regression(self):
        self._train_convert_evaluate_assert(
            allowed_error={"rmse": 0.05, "max_error": 0.81}, subsample=0.5
        )

    def test_boston_housing_simple_decision_tree_regression(self):
        self._train_convert_evaluate_assert(n_estimators=1)

    def test_boston_housing_float_double_corner_case(self):
        self._train_convert_evaluate_assert(
            {
                "colsample_bytree": 1,
                "colsample_bylevel": 1,
                "scale_pos_weight": 1,
                "learning_rate": 0.1,
                "max_delta_step": 0,
                "min_child_weight": 1,
                "n_estimators": 10,
                "subsample": 0.3,
                "objective": "reg:linear",
                "max_depth": 1,
            }
        )

    @pytest.mark.slow
    def test_boston_housing_parameter_stress_test(self):

        options = dict(
            max_depth=[1, 5],
            learning_rate=[0.1, 0.5],
            n_estimators=[1, 10],
            objective=["reg:linear"],
            min_child_weight=[1, 2],
            max_delta_step=[0, 0.1],
            subsample=[1, 0.5, 0.3],
            colsample_bytree=[1, 0.5],
            colsample_bylevel=[1, 0.5],
            scale_pos_weight=[1],
        )

        # Make a cartesian product of all options
        product = itertools.product(*options.values())
        args = [dict(zip(options.keys(), p)) for p in product]

        print("Testing a total of %s cases. This could take a while" % len(args))
        for it, arg in enumerate(args):
            self._train_convert_evaluate_assert(arg)



import keras as _keras
import numpy as _np

_KERAS_LAYERS_1D = [
    _keras.layers.Conv1D,
    _keras.layers.UpSampling1D,
    _keras.layers.ZeroPadding1D,
    _keras.layers.Cropping1D,
    _keras.layers.MaxPooling1D,
    _keras.layers.AveragePooling1D,
    _keras.layers.GlobalMaxPooling1D,
    _keras.layers.GlobalAveragePooling1D,
]

_KERAS_ACTIVATION_LAYERS = [
    _keras.layers.Activation,
    _keras.layers.advanced_activations.LeakyReLU,
    _keras.layers.advanced_activations.PReLU,
    _keras.layers.advanced_activations.ELU,
    _keras.layers.advanced_activations.ThresholdedReLU,
]

_KERAS_NORMALIZATION_LAYERS = [
    _keras.layers.BatchNormalization,
]

_KERAS_RECURRENT_LAYERS = [
    _keras.layers.recurrent.LSTM,
    _keras.layers.recurrent.SimpleRNN,
    _keras.layers.recurrent.GRU,
    _keras.layers.wrappers.Bidirectional,
]

_KERAS_MERGE_LAYERS = [
    _keras.layers.Add,
    _keras.layers.Multiply,
    _keras.layers.Average,
    _keras.layers.Maximum,
    _keras.layers.Concatenate,
    _keras.layers.Dot,
]

_KERAS_SKIP_LAYERS = [
    _keras.layers.core.Dropout,
    _keras.layers.core.SpatialDropout1D,
    _keras.layers.core.SpatialDropout2D,
]

from distutils.version import StrictVersion as _StrictVersion

if _keras.__version__ >= _StrictVersion("2.2.0"):
    from keras.engine.input_layer import InputLayer
else:
    from keras.engine.topology import InputLayer


def _to_list(x):
    if type(x) is not list:
        return [x]
    else:
        return x


def _insert_to_dict(d, key, e):
    # d is a dict where key maps to a list
    if key not in d:
        d[key] = []
    if e not in d[key]:
        d[key].append(e)


def _is_merge_layer(layer):
    for lt in _KERAS_MERGE_LAYERS:
        if isinstance(layer, lt):
            return True
    return False


class NetGraph(object):
    """
    Attributes:
    layer_list - a list of layer names in the Keras model
    connection_map - a map where the key is a layer, the value is a list of its successors
    reverse_connection_map - a map where the key is a layer, the value is a list of its predecessors
    keras_layer_map - a map where the key is a layer name, the value is Keras layer
    model - a reference of the keras model.
    blob_names - blob names for each one of the edge.
    """

    def __init__(self, model):
        self.layer_list = []
        self.edge_map = {}
        self.reverse_edge_map = {}
        self.keras_layer_map = {}

        self.input_layers = []
        self.output_layers = []
        self.layers_inputs = {}  # each layer's input blobs
        self.layers_outputs = {}  # each layer's output blobs

        # these will be pairs of the form (name, shape) because it'll show up on the interface
        self.optional_inputs = []
        self.optional_outputs = []
        self.layers_optional_inputs = {}
        self.layers_optional_outputs = {}

        self.model = model

    def _add_layer(self, keras_layer):
        # add a layer without adding connections.
        # when a layer exist alreday, this operation won't do anything
        layer = keras_layer.name
        if layer not in self.layer_list:
            self.layer_list.append(layer)
            self.keras_layer_map[layer] = keras_layer

    def _replace_blob_name(self, old_name, new_name):
        # replace a blob with a new name
        for l in self.layers_outputs:
            for idx, b in enumerate(self.layers_outputs[l]):
                if b == old_name:
                    self.layers_outputs[l][idx] = new_name

        for l in self.layers_inputs:
            for idx, b in enumerate(self.layers_inputs[l]):
                if b == old_name:
                    self.layers_inputs[l][idx] = new_name

    def get_predecessors(self, layer_name):
        if layer_name in self.reverse_edge_map:
            return self.reverse_edge_map[layer_name][:]  # needs to make a copy
        else:
            return []

    def get_successors(self, layer_name):
        if layer_name in self.edge_map:
            return self.edge_map[layer_name][:]  # needs to make a copy
        else:
            return []

    def get_keras_layer(self, layer_name):
        return self.keras_layer_map[layer_name]

    def get_coreml_layers(self, keras_layer):
        coreml_layers = []
        for key in self.keras_layer_map:
            if self.keras_layer_map[key] == keras_layer:
                coreml_layers.append(key)
        return coreml_layers

    def make_input_layers(self):
        """
        Extract the ordering of the input layers.
        """
        self.input_layers = []
        in_nodes = (
            self.model._inbound_nodes
            if hasattr(self.model, "_inbound_nodes")
            else self.model.inbound_nodes
        )
        if hasattr(self.model, "input_layers"):
            input_keras_layers = self.model.input_layers[:]
            self.input_layers = [None] * len(input_keras_layers)
            for layer in self.layer_list:
                keras_layer = self.keras_layer_map[layer]
                if isinstance(keras_layer, InputLayer):
                    if keras_layer in input_keras_layers:
                        idx = input_keras_layers.index(keras_layer)
                        self.input_layers[idx] = layer
        elif hasattr(self.model, "inputs"):
            for ts in _to_list(self.model.inputs):
                # search for the InputLayer that matches this ts
                for l in self.layer_list:
                    kl = self.keras_layer_map[l]
                    if isinstance(kl, InputLayer) and kl.input == ts:
                        self.input_layers.append(l)
        elif len(in_nodes) <= 1:
            for ts in _to_list(self.model.input):
                # search for the InputLayer that matches this ts
                for l in self.layer_list:
                    kl = self.keras_layer_map[l]
                    if isinstance(kl, InputLayer) and kl.input == ts:
                        self.input_layers.append(l)
        else:
            raise ValueError("Input values cannot be identified.")

    def make_output_layers(self):
        """
        Extract the ordering of output layers.
        """
        self.output_layers = []
        # import pytest; pytest.set_trace()
        if hasattr(self.model, "output_layers"):
            # find corresponding output layers in CoreML model
            # assume output layers are not shared
            # Helper function to recursively extract output layers
            # even if the model has a layer which is a nested model
            def extract_output_layers(keras_model):
                output_layers = []
                for layer in keras_model.output_layers:
                    if hasattr(layer, "output_layers"):
                        output_layers.extend(extract_output_layers(layer))
                    else:
                        output_layers.append(layer)
                return output_layers

            for kl in extract_output_layers(self.model):
                coreml_layers = self.get_coreml_layers(kl)
                if len(coreml_layers) > 0:
                    for cl in coreml_layers:
                        self.output_layers.append(cl)
        elif len(self.model.outputs) > 0:
            for model_output in self.model.outputs:
                for l in self.layer_list:
                    k_layer = self.keras_layer_map[l]
                    in_nodes = (
                        k_layer._inbound_nodes
                        if hasattr(k_layer, "_inbound_nodes")
                        else k_layer.inbound_nodes
                    )
                    for idx in range(len(in_nodes)):
                        out_tensor = k_layer.get_output_at(idx)
                        if out_tensor == model_output or (
                            out_tensor.name in model_output.name
                        ):
                            self.output_layers.append(l)
        if len(self.output_layers) == 0:
            raise ValueError("No outputs can be identified")

    def get_input_layers(self):
        return self.input_layers

    def get_output_layers(self):
        return self.output_layers

    def generate_blob_names(self):
        """
        Generate blob names for each one of the edge.  At this time, Keras does not
        support "fork" operation (a layer with more than 1 blob output). So we just
        use names of the src layer to identify a blob.  We also assume all neural
        networks are singly-connected graphs - which should be the case.
        """
        # generate blob names that represent edges in blob_name_map
        # because of the InputLayers, input blobs are also generated.

        # Generate each layer's input / output blob names
        for layer in self.layer_list:
            keras_layer = self.keras_layer_map[layer]
            # no need to generate InputLayers' blobs
            if not isinstance(keras_layer, InputLayer):
                # layer's input blob names depend on predecessors
                preds = self.get_predecessors(layer)
                for pred in preds:
                    blob_name = pred + "_output"
                    _insert_to_dict(self.layers_inputs, layer, blob_name)
                # layer's output blob is just named after itself
                blob_name = layer + "_output"
                _insert_to_dict(self.layers_outputs, layer, blob_name)

    def get_layer_blobs(self, layer):
        keras_layer = self.keras_layer_map[layer]
        if isinstance(keras_layer, InputLayer):
            return None, None
        else:
            input_blobs = self.layers_inputs[layer]
            output_blobs = self.layers_outputs[layer]
            if layer in self.layers_optional_inputs:
                input_blobs += self.layers_optional_inputs[layer]
            if layer in self.layers_optional_outputs:
                output_blobs += self.layers_optional_outputs[layer]
            return input_blobs, output_blobs

    def reset_model_input_names(self, new_names):
        # call this method after make_input_layers() is called
        if new_names is None:
            return
        if len(new_names) != len(self.input_layers):
            print("Input name length mismatch")
            return
        for i, in_layer in enumerate(self.input_layers):
            old_blob_name = in_layer + "_output"
            new_blob_name = new_names[i]
            succs = self.get_successors(in_layer)
            for succ in succs:
                idx = self.layers_inputs[succ].index(old_blob_name)
                self.layers_inputs[succ][idx] = new_blob_name

    def reset_model_output_names(self, new_names):
        if new_names is None:
            return
        if len(new_names) != len(self.output_layers):
            print("Output name length mismatch")
            return
        for i, out_layer in enumerate(self.output_layers):
            old_blob_name = self.layers_outputs[self.output_layers[i]][0]
            self._replace_blob_name(old_blob_name, new_names[i])

    # need to update both layer's in/out list and graph in/out ports
    def add_recurrent_optionals(self):
        # call this after blob names are generated
        for layer in self.layer_list:
            keras_layer = self.keras_layer_map[layer]
            if type(keras_layer) in _KERAS_RECURRENT_LAYERS:
                if not isinstance(keras_layer, _keras.layers.wrappers.Bidirectional):
                    hidden_size = keras_layer.units
                else:
                    hidden_size = keras_layer.forward_layer.units
                h_in_name = layer + "_h_in"
                h_out_name = layer + "_h_out"
                self.optional_inputs.append((h_in_name, hidden_size))
                self.optional_outputs.append((h_out_name, hidden_size))
                _insert_to_dict(self.layers_optional_inputs, layer, h_in_name)
                _insert_to_dict(self.layers_optional_outputs, layer, h_out_name)
                if isinstance(keras_layer, _keras.layers.recurrent.LSTM):
                    c_in_name = layer + "_c_in"
                    c_out_name = layer + "_c_out"
                    self.optional_inputs.append((c_in_name, hidden_size))
                    self.optional_outputs.append((c_out_name, hidden_size))
                    _insert_to_dict(self.layers_optional_inputs, layer, c_in_name)
                    _insert_to_dict(self.layers_optional_outputs, layer, c_out_name)
                elif isinstance(keras_layer, _keras.layers.wrappers.Bidirectional):
                    c_in_name = layer + "_c_in"
                    c_out_name = layer + "_c_out"
                    h_in_name_rev = layer + "_h_in_rev"
                    c_in_name_rev = layer + "_c_in_rev"
                    h_out_name_rev = layer + "_h_out_rev"
                    c_out_name_rev = layer + "_c_out_rev"
                    self.optional_inputs.append((c_in_name, hidden_size))
                    self.optional_outputs.append((c_out_name, hidden_size))
                    self.optional_inputs.append((h_in_name_rev, hidden_size))
                    self.optional_inputs.append((c_in_name_rev, hidden_size))
                    self.optional_outputs.append((h_out_name_rev, hidden_size))
                    self.optional_outputs.append((c_out_name_rev, hidden_size))
                    _insert_to_dict(self.layers_optional_inputs, layer, c_in_name)
                    _insert_to_dict(self.layers_optional_outputs, layer, c_out_name)
                    _insert_to_dict(self.layers_optional_inputs, layer, h_in_name_rev)
                    _insert_to_dict(self.layers_optional_inputs, layer, c_in_name_rev)
                    _insert_to_dict(self.layers_optional_outputs, layer, h_out_name_rev)
                    _insert_to_dict(self.layers_optional_outputs, layer, c_out_name_rev)

    def _get_first_embedded_model(self):
        for idx, layer in enumerate(self.layer_list):
            keras_layer = self.keras_layer_map[layer]
            if isinstance(keras_layer, _keras.models.Sequential) or isinstance(
                keras_layer, _keras.models.Model
            ):
                return idx
        return -1

    def _get_first_shared_layer(self):
        for idx, layer in enumerate(self.layer_list):
            keras_layer = self.keras_layer_map[layer]
            inbound_nodes = (
                keras_layer.inbound_nodes
                if hasattr(keras_layer, "inbound_nodes")
                else keras_layer._inbound_nodes
            )
            inbound_nodes = [
                node for node in inbound_nodes if len(node.inbound_layers) > 0
            ]
            if (
                not _is_merge_layer(self.keras_layer_map[layer])
                and len(self.get_predecessors(layer)) > 1
                and len(inbound_nodes) > 1
            ):
                return idx
        return -1

    def _get_first_layer_of_type(self, layer_type):
        for idx, layer in enumerate(self.layer_list):
            keras_layer = self.keras_layer_map[layer]
            if isinstance(keras_layer, layer_type):
                return idx
        return -1

    def _add_edge(self, src, snk):
        if src not in self.edge_map:
            self.edge_map[src] = []
        if snk not in self.edge_map[src]:
            self.edge_map[src].append(snk)
        if snk not in self.reverse_edge_map:
            self.reverse_edge_map[snk] = []
        if src not in self.reverse_edge_map[snk]:
            self.reverse_edge_map[snk].append(src)

    def _remove_edge(self, src, snk):
        self.edge_map[src].remove(snk)
        if len(self.edge_map[src]) == 0:
            self.edge_map.pop(src)
        self.reverse_edge_map[snk].remove(src)
        if len(self.reverse_edge_map[snk]) == 0:
            self.reverse_edge_map.pop(snk)

    def _remove_layer(self, layer):
        """
        remove the layer and its input/output edges
        """
        successors = self.get_successors(layer)
        predecessors = self.get_predecessors(layer)
        # remove all edges
        for succ in successors:
            self._remove_edge(layer, succ)
        for pred in predecessors:
            self._remove_edge(pred, layer)
        # remove layer in the data structures
        self.keras_layer_map.pop(layer)
        self.layer_list.remove(layer)

    def _remove_layer_and_reconnect(self, layer):
        """ Remove the layer, and reconnect each of its predecessor to each of
        its successor
        """
        successors = self.get_successors(layer)
        predecessors = self.get_predecessors(layer)
        # remove layer's edges
        for succ in successors:
            self._remove_edge(layer, succ)
        for pred in predecessors:
            self._remove_edge(pred, layer)

        # connect predecessors and successors
        for pred in predecessors:
            for succ in successors:
                self._add_edge(pred, succ)

        # remove layer in the data structures
        self.layer_list.remove(layer)
        self.keras_layer_map.pop(layer)

        # re-assign input and output layers if layer happens to be an
        # input / output layer
        if layer in self.input_layers:
            idx = self.input_layers.index(layer)
            self.input_layers.pop(idx)
            for pred in predecessors:
                self.input_layers.insert(idx, pred)
                idx += 1
        if layer in self.output_layers:
            idx = self.output_layers.index(layer)
            self.output_layers.pop(idx)
            for succ in successors:
                self.output_layers.insert(idx, succ)
                idx += 1

    def _remove_old_edges(self, layer):
        predecessors = self.get_predecessors(layer)
        successors = self.get_successors(layer)
        for pred in predecessors:
            self._remove_edge(pred, layer)
        for succ in successors:
            self._remove_edge(layer, succ)

    def _remove_layers_of_type(self, layer_type):
        idx = self._get_first_layer_of_type(layer_type)
        while idx >= 0:
            layer = self.layer_list[idx]
            self._remove_layer_and_reconnect(layer)
            idx = self._get_first_layer_of_type(layer_type)

    def remove_skip_layers(self, skip_layers):
        for skip_layer in skip_layers:
            self._remove_layers_of_type(skip_layer)

    def remove_internal_input_layers(self):
        idx, nb_layers = 0, len(self.layer_list)
        while idx < nb_layers:
            layer = self.layer_list[idx]
            keras_layer = self.keras_layer_map[layer]
            if (
                isinstance(keras_layer, InputLayer)
                and len(self.get_predecessors(layer)) > 0
            ):
                # these are internal input layers that needs to be taken out
                self._remove_layer_and_reconnect(layer)
                idx -= 1
                nb_layers -= 1
            idx += 1

    def _insert_layer_after(self, layer_idx, new_layer, new_keras_layer):
        """ Insert the new_layer, whose parameter is stored in a Keras layer
        structure new_keras_layer, after the layer whose position is layer_idx.
        """
        layer = self.layer_list[layer_idx]
        self.layer_list.insert(layer_idx + 1, new_layer)
        self.keras_layer_map[new_layer] = new_keras_layer
        successors = self.get_successors(layer)
        # add edge layer -> new_layer
        self._add_edge(layer, new_layer)
        # add edges new_layer -> layer_successor, remove layer -> successor
        for succ in successors:
            self._add_edge(new_layer, succ)
            self._remove_edge(layer, succ)
        # if layer is an output layer, change the output layer tag
        if layer in self.output_layers:
            idx = self.output_layers.index(layer)
            self.output_layers[idx] = new_layer

    def _insert_layer_between(self, src, snk, new_layer, new_keras_layer):
        """ Insert the new_layer, whose keras layer parameters are stored in
        new_keras_layer, between src and snk.
        """
        if snk is None:
            insert_pos = self.layer_list.index(src) + 1
        else:
            insert_pos = self.layer_list.index(snk)  # insert position
        self.layer_list.insert(insert_pos, new_layer)
        self.keras_layer_map[new_layer] = new_keras_layer
        if src is None:  # snk is an input layer
            self._add_edge(new_layer, snk)
        elif snk is None:  # src is an output layer
            self._add_edge(src, new_layer)
        else:
            self._add_edge(src, new_layer)
            self._add_edge(new_layer, snk)
            self._remove_edge(src, snk)

        # if layer is an output layer, change the output layer tag
        if src in self.output_layers:
            idx = self.output_layers.index(src)
            self.output_layers[idx] = new_layer

    def defuse_activation(self):
        """ Defuse the fused activation layers in the network.
        """
        idx, nb_layers = 0, len(self.layer_list)
        while idx < nb_layers:
            layer = self.layer_list[idx]
            k_layer = self.keras_layer_map[layer]
            if isinstance(k_layer, _keras.layers.TimeDistributed):
                k_layer = k_layer.layer
            if (
                isinstance(k_layer, _keras.layers.Conv2D)
                or isinstance(k_layer, _keras.layers.Conv1D)
                or isinstance(k_layer, _keras.layers.SeparableConv2D)
                or isinstance(k_layer, _keras.layers.SeparableConv1D)
                or isinstance(k_layer, _keras.layers.Dense)
            ):

                func_name = (
                    k_layer.activation.__name__
                )

                if func_name != "linear":
                    # Create new layer
                    new_layer = layer + "__activation__"
                    new_keras_layer = _keras.layers.core.Activation(func_name)
                    # insert new layer after it
                    self._insert_layer_after(idx, new_layer, new_keras_layer)
                    idx += 1
                    nb_layers += 1
            idx += 1

    def is_activation(self, layer):
        keras_layer = self.keras_layer_map[layer]
        for activation_type in _KERAS_ACTIVATION_LAYERS:
            if isinstance(keras_layer, activation_type):
                return True
        return False

    def is_1d_layer(self, layer):
        keras_layer = self.keras_layer_map[layer]
        for layer_type in _KERAS_LAYERS_1D:
            if isinstance(keras_layer, layer_type):
                return True
        return False

    def _get_1d_interface_edges(self):
        """ Get edges that represents transition from not-1D to 1D, and 1D to
        not-1D. A 'in_edge e(u,v)' means u operates on non-1D blobs, but v
        operates on 1D blobs. An 'out_edge e(u,v)' means u operates on 1D
        blobs, but v operates on non-1D blobs.
        """
        in_edges = set()
        for layer in self.layer_list:
            if not self.is_1d_layer(layer):
                continue
            preds = self.get_predecessors(layer)
            if len(preds) == 0:
                in_edges.add((None, layer))
            else:
                # because 1D layers are all 1-input,
                # there should only be 1 predecessor
                u, v = preds[0], layer
                while u and (
                    self.is_activation(u) or type(u) in _KERAS_NORMALIZATION_LAYERS
                ):
                    preds = self.get_predecessors(u)
                    v = u
                    u = preds[0] if len(preds) > 0 else None
                if u is None or (not self.is_1d_layer(u)):
                    in_edges.add((u, v))

        out_edges = set()
        for layer in self.layer_list:
            if not self.is_1d_layer(layer):
                continue
            # cases for 1d->output
            if layer in self.output_layers:
                out_edges.add((layer, None))

            succs = self.get_successors(layer)
            if len(succs) > 0:
                # this should be handled in 1d->output already
                if not self.is_activation(succs[0]):
                    for succ in succs:
                        if not self.is_1d_layer(succ):
                            out_edges.add((layer, succ))
                else:
                    act_layer = succs[0]
                    succs = self.get_successors(act_layer)
                    if len(succs) == 0:
                        out_edges.add((act_layer, None))
                    else:
                        for succ in succs:
                            if not self.is_1d_layer(succ):
                                out_edges.add((act_layer, succ))

        return list(in_edges), list(out_edges)

    def insert_1d_permute_layers(self):
        """
        Insert permutation layers before a 1D start point or after 1D end point
        """
        idx, nb_layers = 0, len(self.layer_list)
        in_edges, out_edges = self._get_1d_interface_edges()

        # Hacky Warning: (1) use a 4-D permute, which is not likely to happen in Keras,
        # to represent actual permutation needed for (seq, c, h, w) in CoreML
        # (2) Assume 2-D input shape has meaning (seq, c), and during CoreML runtime,
        # it is represented as 4D blob, (seq, c, h, w)
        for in_edge in in_edges:
            src, snk = in_edge
            if src is None:
                permute_layer = "_permute_" + snk
            else:
                permute_layer = src + "_permute_" + snk
            keras_permute = _keras.layers.Permute(
                dims=(3, 1, 2, 0)
            )  # assume w = 1, switch seq and w
            self._insert_layer_between(src, snk, permute_layer, keras_permute)
        for out_edge in out_edges:
            src, snk = out_edge
            if snk is None:
                permute_layer = src + "_permute_"
            else:
                permute_layer = src + "_permute_" + snk
            keras_permute = _keras.layers.Permute(
                dims=(3, 1, 2, 0)
            )  # assume w = 1, switch seq and w back
            self._insert_layer_between(src, snk, permute_layer, keras_permute)

    def insert_permute_for_spatial_bn(self):

        # find spatial batchnorm layers
        spatial_bn_layers = []
        for layer in self.layer_list:
            keras_layer = self.keras_layer_map[layer]
            if (
                isinstance(keras_layer, _keras.layers.BatchNormalization)
                and len(keras_layer.input_shape) == 4
            ):
                if keras_layer.axis == 1 or keras_layer.axis == 2:
                    spatial_bn_layers.append(layer)

        for sbn in spatial_bn_layers:
            axis = self.keras_layer_map[sbn].axis
            # axis == 1: swap H axis; axis == 2 : swap W axis
            dims = (0, 2, 1, 3) if axis == 1 else (0, 3, 2, 1)
            # add permutation before spatial batchnorm
            pred = self.get_predecessors(sbn)[0]
            permute_layer = pred + "_permute_" + sbn
            keras_permute = _keras.layers.Permute(dims=dims)
            self._insert_layer_between(pred, sbn, permute_layer, keras_permute)
            # add permutation after spatial batchnorm
            succs = self.get_successors(sbn)
            if len(succs) == 0:
                permute_layer = sbn + "_permute_"
                keras_permute = _keras.layers.Permute(dims=dims)
                self._insert_layer_between(sbn, None, permute_layer, keras_permute)
            else:
                for succ in succs:
                    permute_layer = sbn + "_permute_" + succ
                    keras_permute = _keras.layers.Permute(dims=dims)
                    self._insert_layer_between(sbn, succ, permute_layer, keras_permute)

    def build(self, is_top_level=True):
        # sanity check.
        model = self.model
        if not (
            type(model) == _keras.models.Sequential
            or type(model) == _keras.models.Model
        ):
            raise TypeError("Keras layer of type %s is not supported." % type(model))

        # build the graph without considering embedded subgraphs
        for i, layer in enumerate(model.layers):
            in_nodes = (
                layer._inbound_nodes
                if hasattr(layer, "_inbound_nodes")
                else layer.inbound_nodes
            )
            for node in in_nodes:
                for pred in node.inbound_layers:
                    if pred.name not in self.layer_list:
                        self.layer_list.append(pred.name)
                        self.keras_layer_map[pred.name] = pred
                    self._add_edge(pred.name, layer.name)
            self.layer_list.append(layer.name)
            self.keras_layer_map[layer.name] = layer

        # Duplicate models for weight sharing
        idx = self._get_first_shared_layer()
        while idx >= 0:
            layer = self.layer_list[idx]
            keras_layer = self.keras_layer_map[layer]
            predecessors = self.reverse_edge_map[layer]
            successors = self.edge_map[layer]
            new_layers = [layer + "_" + str(i) for i in range(len(predecessors))]
            self.layer_list[idx : idx + 1] = new_layers
            for i, new_layer in enumerate(new_layers):
                self.edge_map[new_layer] = []
                self.reverse_edge_map[new_layer] = []
                self.keras_layer_map[new_layer] = keras_layer
                pred = predecessors[i]
                self._add_edge(pred, new_layer)
                for succ in successors:
                    self._add_edge(new_layer, succ)
            self._remove_old_edges(layer)
            self.keras_layer_map.pop(layer)
            idx = self._get_first_shared_layer()

        # Expand the sub-models
        idx = self._get_first_embedded_model()
        while idx >= 0:
            # grab the input and output edges of the embedded model
            embedded_model = self.layer_list[idx]
            # build the embedded model
            embedded_keras_model = self.keras_layer_map[embedded_model]
            embedded_graph = NetGraph(embedded_keras_model)
            embedded_graph.build(is_top_level=False)
            # replace the embedded model with the layers of the embedded graph
            embedded_layer_list = embedded_graph.layer_list
            new_layer_list = []
            for embedded_layer_name in embedded_layer_list:
                new_layer_name = embedded_model + "_" + embedded_layer_name
                new_layer_list.append(new_layer_name)
                self.keras_layer_map[new_layer_name] = embedded_graph.keras_layer_map[
                    embedded_layer_name
                ]
                # add edge [embed_layer -> its succ]
                embedded_successors = embedded_graph.get_successors(embedded_layer_name)
                for embed_succ_name in embedded_successors:
                    new_embed_succ_name = embedded_model + "_" + embed_succ_name
                    self._add_edge(new_layer_name, new_embed_succ_name)
                # add edge [pred -> embed_layer]
                embedded_predecessors = embedded_graph.get_predecessors(
                    embedded_layer_name
                )
                for embed_pred_name in embedded_predecessors:
                    new_embed_pred_name = embedded_model + "_" + embed_pred_name
                    self._add_edge(new_embed_pred_name, new_layer_name)

            self.layer_list[idx + 1 : idx + 1] = new_layer_list
            # replace input / output edges to the model with input/output edges of the embedded layers
            predecessors = self.get_predecessors(embedded_model)
            embedded_inputs = embedded_graph.get_input_layers()
            for i, pred in enumerate(predecessors):
                embed_input = embedded_inputs[i]
                new_embed_input = embedded_model + "_" + embed_input
                self._add_edge(pred, new_embed_input)

            embedded_outputs = embedded_graph.get_output_layers()
            successors = self.get_successors(embedded_model)
            for i, succ in enumerate(successors):
                embed_output = embedded_outputs[i]
                new_embed_output = embedded_model + "_" + embed_output

                self._add_edge(new_embed_output, succ)

            # clear up the embedded model
            self._remove_layer(embedded_model)
            idx = self._get_first_embedded_model()

        # tag input layers and and output layers
        self.make_input_layers()
        self.make_output_layers()

        # make graph level adjustments - do this only on top level
        if is_top_level:
            self.remove_skip_layers(_KERAS_SKIP_LAYERS)  # done 1 pass
            self.insert_1d_permute_layers()
            self.insert_permute_for_spatial_bn()
            self.defuse_activation()
            self.remove_internal_input_layers()

    def print_layer_list(self):
        print("\n")
        print("layer_list")
        print(self.layer_list)

    def print_edge_map(self):
        print("\n")
        print("edge map:")
        for src in self.edge_map:
            for snk in self.edge_map[src]:
                print("  ", src, "-->", snk)

    def print_reverse_edge_map(self):
        print("\n")
        print("reverse edge map: ")
        for snk in self.reverse_edge_map:
            for src in self.reverse_edge_map[snk]:
                print("  ", snk, "<--", src)

    def print_mapping(self):
        print("\nmapping:")
        for key in self.keras_layer_map:
            print(
                key,
                "-->",
                self.keras_layer_map[key],
                "(",
                self.keras_layer_map[key].name,
                ")",
            )

    def print_all(self):
        print("=" * 80)
        self.print_layer_list()
        self.print_edge_map()
        self.print_reverse_edge_map()
        self.print_mapping()



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
from sklearn.ensemble import RandomForestRegressor
from coremltools._deps import _HAS_SKLEARN
from coremltools.proto import Model_pb2
from coremltools.proto import FeatureTypes_pb2

if _HAS_SKLEARN:
    from sklearn.ensemble import RandomForestRegressor
    from coremltools.converters import sklearn as skl_converter


@unittest.skipIf(not _HAS_SKLEARN, "Missing scikit-learn. Skipping tests.")
class RandomForestRegressorScikitTest(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        from sklearn.datasets import load_boston
        from sklearn.ensemble import RandomForestRegressor

        scikit_data = load_boston()
        scikit_model = RandomForestRegressor(random_state=1)
        scikit_model.fit(scikit_data["data"], scikit_data["target"])

        # Save the data and the model
        self.scikit_data = scikit_data
        self.scikit_model = scikit_model

    def test_conversion(self):
        input_names = self.scikit_data.feature_names
        output_name = "target"
        spec = skl_converter.convert(
            self.scikit_model, input_names, "target"
        ).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)

        # Test the interface class
        self.assertEquals(spec.description.predictedFeatureName, "target")

        # Test the inputs and outputs
        self.assertEquals(len(spec.description.output), 1)
        self.assertEquals(spec.description.output[0].name, "target")
        self.assertEquals(
            spec.description.output[0].type.WhichOneof("Type"), "doubleType"
        )
        for input_type in spec.description.input:
            self.assertEquals(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )

        # Test the linear regression parameters.
        self.assertEquals(len(spec.pipelineRegressor.pipeline.models), 2)
        tr = spec.pipelineRegressor.pipeline.models[
            -1
        ].treeEnsembleRegressor.treeEnsemble
        self.assertIsNotNone(tr)
        self.assertEquals(len(tr.nodes), 5996)

    def test_conversion_bad_inputs(self):
        # Error on converting an untrained model
        with self.assertRaises(Exception):
            model = RandomForestRegressor()
            spec = skl_converter.convert(model, "data", "out")

        # Check the expected class during covnersion.
        from sklearn.preprocessing import OneHotEncoder

        with self.assertRaises(Exception):
            model = OneHotEncoder()
            spec = skl_converter.convert(model, "data", "out")



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _

import logging
import os
import gc

import tensorflow as tf

from tempfile import mktemp
from .basic_graph_ops import fill_outputs
from .converter import TFConverter
from .tf_graph_pass import *  # pylint: disable=unused-wildcard-import,wildcard-import
from .tfssa import NetworkEnsemble, SSAFunction
from .parsed_tf_node import ParsedTFNode
from coremltools.converters._profile_utils import _profile
from tqdm import tqdm as _tqdm
from distutils.version import StrictVersion as _StrictVersion
from coremltools._deps import __get_version as _get_version


class TFLoader:
    """Abstract class for TensorFlow model loader."""

    def __init__(self, model, debug=False, **kwargs):
        """
        TensorFlow model loader.

        Parameters
        ----------
        model: TensorFlow model
            Model generated using TensorFlow.
        debug: bool, optional, defaults to False
            If true, display verbose logging and visualizations.
        kwargs: dict(str, Any), optional, defaults to None
            Dictionary of additional arguments.
        """
        self.model = model
        self.debug = debug
        self.kwargs = kwargs
        self._graph_def = None
        self._tf_ssa = None

    @_profile
    def load(self):
        """Load TensorFlow model into MIL program."""

        logging.info("Loading TensorFlow model '{}'".format(self.model))
        outputs = self.kwargs.get("outputs", None)
        self._graph_def = self._graph_def_from_model(outputs)

        if self._graph_def is not None and len(self._graph_def.node) == 0:
            msg = "tf.Graph should have at least 1 node, Got empty graph."
            raise ValueError(msg)

        self._tf_ssa = self._tf_ssa_from_graph_def()

        del self._graph_def
        gc.collect()

        if self.debug:
            import graphviz

            dot_string = self._tf_ssa.get_dot_string(
                annotation=True, name_and_op_style=True, highlight_debug_nodes=[]
            )
            graphviz.Source(dot_string).view(
                filename="/tmp/ssa_before_tf_passes", cleanup=True
            )

        program = self._program_from_tf_ssa()
        logging.debug("program:\n{}".format(program))
        return program

    # @abstractmethod
    def _graph_def_from_model(self, outputs=None):
        """Load TensorFlow model into GraphDef. Overwrite for different TF versions."""
        pass

    # @abstractmethod
    def _tf_ssa_from_graph_def(self, fn_name="main"):
        """Load GraphDef and parse into NetworkEnsemble (TFSSA)."""
        pass

    # @abstractmethod
    def _program_from_tf_ssa(self):
        """Load NetworkEnsemble (TFSSA) and parse into MIL program."""
        pass

    @staticmethod
    def extract_sub_graph(graph_def, outputs=None):
        """Extract sub-graph based on user-provided outputs."""
        if outputs is None or len(outputs) == 0:
            return graph_def
        msg = "Extracting sub-graph based on outputs '{}' from the full model"
        logging.debug(msg.format(outputs))
        outputs = outputs if isinstance(outputs, list) else [outputs]
        outputs = [i.split(":")[0] for i in outputs]
        if _get_version(tf.__version__) < _StrictVersion("1.13.1"):
            return tf.graph_util.extract_sub_graph(graph_def, outputs)
        else:
            return tf.compat.v1.graph_util.extract_sub_graph(graph_def, outputs)


class TF1Loader(TFLoader):
    def __init__(self, model, debug=False, **kwargs):
        """
        TensorFlow 1.x model loader.

        Parameters
        ----------
        model: Model created with TensorFlow 1.x
            One of the following model format:
                - TensorFlow tf.Graph object or frozen graph (.pb) file path
                - TensorFlow tf.keras.Model object or HDF5 (.h5) file path
                - TensorFlow SavedModel directory path
        debug: bool, optional. Defaults to False.
            This flag should generally be False except for debugging purposes
            for diagnosing conversion errors. Setting this flag to True will
            cause graph pass errors to be ignored, forcefully returning a
            NetworkEnsemble object.
        kwargs: dict(str, Any), optional
            Dictionary of additional arguments.
        """
        TFLoader.__init__(self, model, debug, **kwargs)

    def _graph_def_from_model(self, outputs=None):
        """Overwrites TFLoader._graph_def_from_model()"""
        msg = "Expected model format: [tf.Graph | .pb | SavedModel | tf.keras.Model | .h5], got {}"
        if isinstance(self.model, tf.Graph) and hasattr(self.model, "as_graph_def"):
            graph_def = self.model.as_graph_def(add_shapes=True)
            return self.extract_sub_graph(graph_def, outputs)
        elif isinstance(self.model, tf.keras.Model):
            graph_def = self._from_tf_keras_model(self.model)
            return self.extract_sub_graph(graph_def, outputs)
        elif isinstance(self.model, str):
            if not os.path.exists(str(self.model)):
                raise ValueError('Input model "{}" does not exist'.format(self.model))
            elif os.path.isfile(str(self.model)) and self.model.endswith(".pb"):
                if _get_version(tf.__version__) < _StrictVersion("1.13.1"):
                    with open(self.model, "rb") as f:
                        gd = tf.GraphDef()
                        gd.ParseFromString(f.read())
                    with tf.Graph().as_default() as graph:
                        tf.import_graph_def(gd, name="")
                else:
                    with tf.io.gfile.GFile(self.model, "rb") as f:
                        gd = tf.compat.v1.GraphDef()
                        gd.ParseFromString(f.read())
                    with tf.Graph().as_default() as graph:
                        tf.graph_util.import_graph_def(gd, name="")
                graph_def = graph.as_graph_def(add_shapes=True)
                return self.extract_sub_graph(graph_def, outputs)
            elif os.path.isfile(str(self.model)) and self.model.endswith(".h5"):
                graph_def = self._from_tf_keras_model(self.model)
                return self.extract_sub_graph(graph_def, outputs)
            elif os.path.isdir(str(self.model)):
                graph_def = self._from_saved_model(self.model)
                return self.extract_sub_graph(graph_def, outputs)
            else:
                raise NotImplementedError(msg.format(self.model))
        else:
            raise NotImplementedError(msg.format(self.model))

    def _tf_ssa_from_graph_def(self, fn_name="main"):
        """Overwrites TFLoader._tf_ssa_from_graph_def()"""
        graph_dict = {}
        for node in self._graph_def.node:
            graph_dict[node.name] = ParsedTFNode(node)

        tensor_array_resource_removal(graph_dict)
        graph = insert_get_tuple(graph_dict)
        graph = fill_outputs(graph)
        delete_disconnected_nodes(graph)

        tf_ssa = NetworkEnsemble()
        tf_ssa.functions[fn_name] = SSAFunction(graph)
        return tf_ssa

    def _program_from_tf_ssa(self):
        """Overwrites TFLoader._mil_program_from_tf_ssa()"""
        # Applying frontend passes on TFSSA. Note that these are different from
        # passes applied to MIL in TF frontend.
        tf_passes = [
            delete_asserts,
            functionalize_loops,
            constant_propagation,
            quantization_pass,
            cond_to_where,
            remove_variable_nodes,
            fuse_dilation_conv,
        ]

        if self.debug:
            for tf_pass in _tqdm(
                tf_passes, desc="Running TensorFlow Graph Passes", unit=" passes"
            ):
                try:
                    tf_pass(self._tf_ssa)
                except Exception as e:
                    logging.exception('Exception in pass "{}": {}'.format(tf_pass, e))
                    logging.info("Ignoring exception and continuing to next pass")
        else:
            for tf_pass in _tqdm(
                tf_passes, desc="Running TensorFlow Graph Passes", unit=" passes"
            ):
                tf_pass(self._tf_ssa)

        if self.debug:
            import graphviz

            dot_string = self._tf_ssa.get_dot_string(
                annotation=True, name_and_op_style=True, highlight_debug_nodes=[]
            )
            graphviz.Source(dot_string).view(
                filename="/tmp/ssa_after_tf_passes", cleanup=True
            )

        converter = TFConverter(self._tf_ssa, **self.kwargs)
        return converter.convert()

    @staticmethod
    def _from_saved_model(saved_model_dir):
        from tensorflow.python.tools import freeze_graph

        # must import here as tf.contrib is only available on TF 1.x
        from tensorflow.contrib.saved_model.python.saved_model import reader

        saved_model_tags = reader.get_saved_model_tag_sets(saved_model_dir)[0]
        if not saved_model_tags:
            msg = "Unsupported SavedModel directory format: no tag_sets available"
            raise NotImplementedError(msg)

        # get model outputs
        output_node_names = []
        if _get_version(tf.__version__) < _StrictVersion("1.13.1"):
            sess = tf.Session()
        else:
            sess = tf.compat.v1.Session()
        metagraph = tf.saved_model.loader.load(
            sess, saved_model_tags, saved_model_dir
        )
        for sd in metagraph.signature_def.values():
            output_node_names += [o.name.split(":")[0] for o in sd.outputs.values()]

        sess.close()

        # get frozen graph
        output_graph = mktemp()
        tf.compat.v1.reset_default_graph() if _get_version(tf.__version__) >= _StrictVersion("1.13.1") else tf.reset_default_graph()
        freeze_graph.freeze_graph(
            input_graph=None,
            input_saver=None,
            input_binary=None,
            input_checkpoint=None,
            output_node_names=",".join(output_node_names),
            restore_op_name=None,
            filename_tensor_name=None,
            output_graph=output_graph,
            clear_devices=True,
            initializer_nodes="",
            variable_names_whitelist="",
            variable_names_blacklist="",
            input_meta_graph=None,
            input_saved_model_dir=saved_model_dir,
            saved_model_tags=",".join(saved_model_tags),
        )

        if _get_version(tf.__version__) < _StrictVersion("1.13.1"):
            graph_def = tf.GraphDef()
            with open(output_graph, "rb") as f:
                graph_def.ParseFromString(f.read())
            graph_def = tf.graph_util.remove_training_nodes(graph_def)
        else:
            graph_def = tf.compat.v1.GraphDef()
            with open(output_graph, "rb") as f:
                graph_def.ParseFromString(f.read())
            graph_def = tf.compat.v1.graph_util.remove_training_nodes(graph_def)
        with tf.Graph().as_default() as graph:
            tf.graph_util.import_graph_def(graph_def, name="")
        return graph.as_graph_def(add_shapes=True)

    @staticmethod
    def _from_tf_keras_model(keras_model):
        from tensorflow.python.keras.saving import saving_utils
        from tensorflow.python.framework.convert_to_constants import (
            convert_variables_to_constants_v2,
        )

        if not isinstance(keras_model, tf.keras.Model):
            keras_model = tf.keras.models.load_model(keras_model, None)

        tf.keras.backend.clear_session()
        tf.keras.backend.set_learning_phase(False)
        fn = saving_utils.trace_model_call(keras_model)
        cf = fn.get_concrete_function()
        try:
            frozen_fn = convert_variables_to_constants_v2(cf)
            return frozen_fn.graph.as_graph_def(add_shapes=True)
        except Exception:
            raise NotImplementedError("Unhandled tf.keras model format")



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from .activation import *
from .control_flow import *
from .conv import *
from .elementwise_binary import *
from .elementwise_unary import *
from .image_resizing import *
from .linear import *
from .normalization import *
from .pool import *
from .random import *
from .recurrent import *
from .reduction import *
from .scatter_gather import *
from .tensor_operation import *
from .tensor_transformation import *



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import numpy as np
import torch
import torch.nn as nn
from coremltools import TensorType, RangeDim
import coremltools.models.utils as coremltoolsutils
from ..converter import torch_to_mil_types
from coremltools.models import MLModel
from coremltools._deps import _IS_MACOS, _HAS_TORCH
from coremltools.converters.mil.mil.types.type_mapping import nptype_from_builtin
from coremltools.converters.mil.testing_reqs import ct
import pytest


class ModuleWrapper(nn.Module):
    """
    Helper class to transform torch function into torch nn module.
    This helps to keep the testing interface same for torch functional api.

    """
    def __init__(self, function, kwargs=None):
        super(ModuleWrapper, self).__init__()
        self.function = function
        self.kwargs = kwargs if kwargs else {}

    def forward(self, *args):
        return self.function(*args, **self.kwargs)


np.random.seed(1984)


def _flatten(object):
    flattened_list = []
    for item in object:
        if isinstance(item, (list, tuple)):
            flattened_list.extend(_flatten(item))
        else:
            flattened_list.append(item)
    return flattened_list


def contains_op(torch, op_string):
    if hasattr(torch, op_string):
        return True
    return False


def convert_to_coreml_inputs(input_description, inputs):
    """Convenience function to combine a CoreML model's input description and
    set of raw inputs into the format expected by the model's predict function.
    """
    flattened_inputs = _flatten(inputs)
    coreml_inputs = {
        str(x): inp.numpy().astype(np.float32) for x, inp in zip(input_description, flattened_inputs)
    }

    for k, v in coreml_inputs.items():
        if isinstance(v, np.ndarray) and v.ndim == 0:
            coreml_inputs[k] = np.expand_dims(v, axis=-1)

    return coreml_inputs


def convert_to_mlmodel(model_spec, tensor_inputs, backend="nn_proto"):
    def _convert_to_inputtype(inputs):
        if isinstance(inputs, list):
            return [_convert_to_inputtype(x) for x in inputs]
        elif isinstance(inputs, tuple):
            return tuple([_convert_to_inputtype(x) for x in inputs])
        elif isinstance(inputs, TensorType):
            return inputs
        elif isinstance(inputs, torch.Tensor):
            return TensorType(shape=inputs.shape, dtype=torch_to_mil_types[inputs.dtype])
        else:
            raise ValueError(
                "Unable to parse type {} into InputType.".format(type(inputs))
            )

    inputs = list(_convert_to_inputtype(tensor_inputs))
    return ct.convert(model_spec, inputs=inputs, convert_to=backend,
        source="pytorch")


def generate_input_data(input_size, rand_range=(0, 1)):
    r1, r2 = rand_range

    def random_data(spec):
        if isinstance(spec, TensorType):
            spec_shape = spec.shape.shape
            dtype = nptype_from_builtin(spec.dtype)
        else:
            spec_shape = spec
            dtype = np.float32

        static_shape = tuple([np.random.randint(dim.lower_bound, dim.upper_bound if dim.upper_bound > 0 else 10)
                              if isinstance(dim, RangeDim) else dim for dim in spec_shape])

        data = np.random.rand(*static_shape) if static_shape != () else np.random.rand()
        data = (r1 - r2) * data + r2
        return torch.from_numpy(np.array(data).astype(dtype))

    if isinstance(input_size, list):
        return [random_data(size) for size in input_size]
    else:
        return random_data(input_size)


def trace_model(model, input_data):
    model.eval()
    if isinstance(input_data, list):
        input_data = tuple(input_data)
    torch_model = torch.jit.trace(model, input_data)
    return torch_model


def flatten_and_detach_torch_results(torch_results):
    if isinstance(torch_results, (list, tuple)):
        return [x.detach().numpy() for x in _flatten(torch_results) if x is not None]
    # Do not need to flatten
    return [torch_results.detach().numpy()]


def convert_and_compare(input_data, model_spec,
        expected_results=None, atol=1e-5,
        backend="nn_proto"):
    """
    If expected results is not set, it will by default
    be set to the flattened output of the torch model.

    Inputs:

    - input_data: torch.tensor or list[torch.tensor]
    """
    if isinstance(model_spec, str):
        torch_model = torch.jit.load(model_spec)
    else:
        torch_model = model_spec

    if not isinstance(input_data, (list, tuple)):
        input_data = [input_data]

    if not expected_results:
        expected_results = torch_model(*input_data)
    expected_results = flatten_and_detach_torch_results(expected_results)
    mlmodel = convert_to_mlmodel(model_spec, input_data, backend=backend)
    coreml_inputs = convert_to_coreml_inputs(mlmodel.input_description,
                                             input_data)
    if _IS_MACOS:
        if not coremltoolsutils._has_custom_layer(mlmodel.get_spec()):
            coreml_results = mlmodel.predict(coreml_inputs, useCPUOnly=True)
            sorted_coreml_results = [
                coreml_results[key] for key in sorted(coreml_results.keys())
            ]
            for torch_result, coreml_result in zip(expected_results,
                                                   sorted_coreml_results):
                np.testing.assert_equal(coreml_result.shape, torch_result.shape)
                np.testing.assert_allclose(coreml_result, torch_result,
                                           atol=atol)
            return model_spec, mlmodel, coreml_inputs, coreml_results


class TorchBaseTest(object):
    testclassname=''
    testmodelname=''
    @pytest.fixture(autouse=True)
    def store_testname_with_args(self, request):
        TorchBaseTest.testclassname = type(self).__name__
        TorchBaseTest.testmodelname = request.node.name

    def teardown_method(self, method):
        pass

    @staticmethod
    def run_compare_torch(
            input_data, model, expected_results=None, places=5,
            input_as_shape=True, backend="nn_proto",
            rand_range=(0.0, 1.0), use_scripting=False,
    ):
        """
            Traces a model and runs a numerical test.
            Args:
                input_as_shape <bool>: If true generates random input data with shape.
                expected_results <iterable, optional>: Expected result from running pytorch model.
        """
        model.eval()
        if input_as_shape:
            input_data = generate_input_data(input_data, rand_range)
        model_spec = torch.jit.script(model) if use_scripting else trace_model(
            model, input_data)
        model_spec, mlmodel, coreml_inputs, coreml_results = \
            convert_and_compare(
            input_data, model_spec, expected_results=expected_results,
            atol=10.0 ** -places, backend=backend,
        )

        return model_spec, mlmodel, coreml_inputs, coreml_results, \
               TorchBaseTest.testclassname, TorchBaseTest.testmodelname



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

# expose directories as imports
from . import libsvm
from . import sklearn
from . import xgboost
from . import keras
from . import caffe
from . import onnx
from ._converters_entry import convert
from .mil import (
    ClassifierConfig,
    TensorType,
    ImageType,
    RangeDim,
    Shape,
    EnumeratedShapes,
)
